
import logging
import syslog
import sys
import signal
import alsaaudio
import subprocess
import json
from urllib.parse import urlparse

from uci import Uci

from barix.media.SipClientAudioSource import *
from barix.media.RtpAudioSource import *
from barix.web.system_info import getMACaddr

from device import Device
from ms700 import DeviceMs700
from fw_updater import FwUpdater
from uci_util import uciGet
from tcp_api import TcpApi
from playback_manager import PlaybackManager
from inout import InOut


class EpicApp:
    sipDict = {
        SipCallState.IDLE: "IDLE",
        SipCallState.CALLING: "CALLING",
        SipCallState.INCOMMING_CALL: "INCOMING",
        SipCallState.RINGING: "RINGING",
        SipCallState.CALLING_RINGING: "RINGING",
        SipCallState.CONNECTED: "CONNECTED",
        SipCallState.GETTING_NOTIF: "GETTING_NOTIF"
    }

    _instance   = None
    _device=None
    _deviceIp="127.0.0.1"
    _buttonStatus=0x0F
    _buttonPStatus=0x0F
    _prev_hb_time=0

    @classmethod
    def setup(cls, device: Device) -> None:
        cls._instance   = EpicApp(device)

    @classmethod
    def get(cls) -> "EpicApp":
        return cls._instance

    def __init__(self, device: Device):
        self.logger = logging.getLogger("epic")
        with open("/barix/info/VERSION", "r") as fIn:
            self.sysVersion = fIn.readline().strip()

        self.device = device
        self.model = self.device.getModel()

        syslog.openlog("AE-MSx00")
        msg = "MS-x00 controller starting. model={} sysVersion={}".format(self.model, self.sysVersion)
        self.logger.info(msg)
        syslog.syslog(msg)
        self.stopApplication = False
        signal.signal(signal.SIGTERM, self.signalHandler)
        signal.signal(signal.SIGINT, self.signalHandler)
        signal.signal(signal.SIGUSR2, self.signalHandler)
        self._nwkAudioActive = None
        self._playingNotification = None
        self.amp_mute_status=-1

    def setRelay1(self, stt: bool) -> None:
        """
        Sets the state of relay1 to ON(True) or OFF(False)

        Returns true is state changed, which requires a status
        update to be sent through TCP connected clients
        """
        if stt:
            sttStr = "ON"
        else:
            sttStr = "OFF"

        syslog.syslog("RELAY1 --> {}".format(sttStr))
        self.device.setRelay(1, stt)

    def setRelay2(self, stt: bool) -> None:
        """
        Sets the state of relay2 to ON(True) or OFF(False)

        Returns true is state changed, which requires a status
        update to be sent through TCP connected clients
        """
        if stt:
            sttStr = "ON"
        else:
            sttStr = "OFF"

        syslog.syslog("RELAY2 --> {}".format(sttStr))
        self.device.setRelay(2, stt)

    def signalHandler(self, sigNum: int, frame):
        msg = "caught signal {}".format(sigNum)
        syslog.syslog(msg)
        self.logger.info(msg)
        if sigNum==signal.SIGUSR2:
            msg = "activate display"
            syslog.syslog(msg)
            self.logger.info(msg)
            self.device.activateDisplay()
        else:
            msg = "stop the application"
            syslog.syslog(msg)
            self.logger.info(msg)
            self.stopApplication = True

    def refreshIP(self):
        self.device.refreshIp()

    def reloadSettings(self):
        logger  = logging.getLogger("epic")
        logger.info("Reloading settings...")
        try:
            self.device.reloadSettings()
        except Exception as e:
            logger.error("Error reloading settings: ", exc_info=e)

    def enableSSH(self, state: bool):
        self.device.enableSSH(state)

    def run(self):
        ret_val=0
        try:
            uci = Uci()
            fwUrl = uciGet(uci, "epic", "main", "fw_update_url")
            fwAutoUpdate = uciGet(uci, "epic", "main", "fw_update_auto")
            if fwUrl is not None and fwAutoUpdate == "true":
                self.fwUpdater = FwUpdater(fwUrl, "/mnt/data/state/epic-sip/fw_last_modified.conf")
                self.fwUpdater.setIniWindowTime(900)
                self.fwUpdater.setUpdateBaseTime(2, 0)
                self.fwUpdater.setUpdateWindowDuration(120)
            else:
                self.fwUpdater = None

            self.device.setRelay(1, False)
            self.device.setRelay(2, False)
            self.device.setRedLed(False)
            self.device.setGreenLed(False)
            # self.redLed         = leds.Led("barix:led1:red")
            # self.greenLed       = leds.Led("barix:led1:green")
            # self.greenLed.set(True)
            # self.redLed.set(False)
            syslog.syslog("RED LED --> OFF")
            syslog.syslog("GREEN LED --> OFF")

#            self.mainVolume = int(uci.get("epic", "main", "volume"))
            self.mainVolume = int(uci.get("ae_ms","volume","out0"))
            if self.mainVolume > 16:
                self.mainVolume = 16
            elif self.mainVolume < -79:
                self.mainVolume = -79

            self.device.loadMixers()

            tcpApiPort = int(uciGet(uci, "epic", "api", "abcl_tcp_port"))
            if tcpApiPort != 0:
                self.tcpApi = TcpApi("0.0.0.0", tcpApiPort)
                self.tcpApi.setAbclHandler(self.executeAbcl)
                self.tcpApi.launch()
            else:
                self.tcpApi = None

            self.playbackManager = PlaybackManager()

            bgmUrl = uciGet(uci, "epic", "bgm", "url")
            bgmBufSz = int(uci.get("epic", "bgm", "buffer_size"))
            bgmVol = int(uci.get("epic", "bgm", "volume"))
            if bgmUrl is not None and len(bgmUrl) > 0:
                urlComps = urlparse(bgmUrl)
                if urlComps.scheme == "rtp":
                    bgmParams = dict()
                    bgmParams["url"] = bgmUrl
                    bgmParams["playback_mode"] = "full_buffer"
#                    bgmParams["playback_mode"] = "const_delay"     # UCI param epic.bgm.buffer_size should be set to 50
                    bgmParams["input_buffer_size"] = 0
                    bgmParams["output_buffer_size"] = bgmBufSz
                    bgmSrc = RtpAudioSource(1, "bgm", bgmVol, bgmParams, self.device.getRtpBgmCh(), 48000)
                    logMsg = "BGM SOURCE: {} buff={}ms vol={}%".format(bgmUrl, bgmBufSz, bgmVol)
                    self.logger.info(logMsg)
                    syslog.syslog(logMsg)
                    self.playbackManager.setBgmSource(bgmSrc)
                else:
                    self.logger.error("BGM url is not supported: {}".format(bgmUrl))

            hpnUrl = uciGet(uci, "epic", "hpn", "url")

            if hpnUrl is not None and len(hpnUrl) > 0:
                hpnBufSz = int(uci.get("epic", "hpn", "buffer_size"))
                hpnVol = int(uci.get("epic", "hpn", "volume"))
                if uci.get("epic", "hpn", "high_prio").lower() == "true":
                    self.hpnHighPrio = True
                else:
                    self.hpnHighPrio = False

                urlComps = urlparse(hpnUrl)
                if urlComps.scheme == "rtp":
                    hpnParams = dict()
                    hpnParams["url"] = hpnUrl
                    hpnParams["playback_mode"] = "full_buffer"
#                    hpnParams["playback_mode"] = "const_delay"     # UCI param epic.hpn.buffer_size should be set to 50
                    hpnParams["input_buffer_size"] = 0
                    hpnParams["output_buffer_size"] = hpnBufSz
                    hpnSrc = RtpAudioSource(2, "hpn", hpnVol, hpnParams, self.device.getRtpHpnCh(), 48000)
                    self.playbackManager.setHpnSource(hpnSrc)
                    logMsg = "HPN SOURCE: {} buff={}ms vol={}%".format(hpnUrl, hpnBufSz, hpnVol)
                    self.logger.info(logMsg)
                    syslog.syslog(logMsg)
                    self.playbackManager.setHpnHighPrio(self.hpnHighPrio)
                else:
                    self.logger.error("HPN url is not supported: {}".format(hpnUrl))

            self.autodial1 = uciGet(uci, "epic", "sip", "autodial1")
            self.autodial2 = uciGet(uci, "epic", "sip", "autodial2")
            self.autodial3 = uciGet(uci, "epic", "sip", "autodial3")
            self.autodial4 = uciGet(uci, "epic", "sip", "autodial4")

            sipEnable       = uciGet(uci, "epic", "sip", "enabled")
            self.audioInput = uciGet(uci, "epic", "sip", "input")
            self.micInGain  = uciGet(uci, "epic", "sip", "micin_gain")
            self.lineInGain = uciGet(uci, "epic", "sip", "linein_gain")
            sipVol          = int(uci.get("epic", "sip", "volume"))
            sipDomain       = uciGet(uci, "epic", "sip", "domain")
            self.sipUserName= uciGet(uci, "epic", "sip", "username")
            sipPassword     = uciGet(uci, "epic", "sip", "password")
            autoAnswer      = uciGet(uci, "epic", "sip", "auto_answer")
            answerTime      = uciGet(uci, "epic", "sip", "answer_time")
            callTimeout     = uciGet(uci, "epic", "sip", "call_timeout")
            streamTimeout   = uciGet(uci, "epic", "sip", "stream_timeout")
            sipTransport    = uciGet(uci, "epic", "sip", "transport")
            beepOnAnswer    = uciGet(uci, "epic", "sip", "beep_on_answer")
            halfDuplex      = uciGet(uci, "epic", "sip", "half_duplex")
            aecEnabled      = uciGet(uci, "epic", "sip", "aec")
            inputGain       = 0.01 * int(uciGet(uci, "epic", "sip", "input_gain"))
            hdxLevel        = float(uciGet(uci, "epic", "sip", "hdx_level"))
            hdxTimeout      = 0.001 * float(uciGet(uci, "epic", "sip", "hdx_timeout"))

            #self.setAudioInputSource()

            self.sipSrc = None
            if sipEnable == "true":
                if len(sipDomain)>0 and len(self.sipUserName)>0 and len(sipPassword)>0:
                    sipParams = dict()
                    sipParams["sip_mode"] = "server"
                    sipParams["domain"] = sipDomain
                    sipParams["user_name"] = self.sipUserName
                    sipParams["password"] = sipPassword
                    sipParams["transport_and_encryption"] = sipTransport
                    sipParams["auto_answer"] = autoAnswer
                    sipParams["answer_time"] = answerTime
                    sipParams["call_timeout"] = callTimeout
                    sipParams["stream_timeout"] = streamTimeout
                    sipParams["beep_on_answer"] = beepOnAnswer
                    sipParams["half_duplex"] = halfDuplex
                    sipParams["aec"] = aecEnabled
                    sipParams["input_gain"] = inputGain
                    sipParams["hdx_level"] = hdxLevel
                    sipParams["hdx_timeout"] = hdxTimeout
                    self.logger.info("loading SIP with '{}'".format(sipParams))
                    self.sipSrc = SipClient(3, "sip", sipVol, sipParams, self.device.getSipPcmp(),
                                            self.device.getSipPcmc())
                    self.playbackManager.setSipSource(self.sipSrc)
                    syslog.syslog("Loaded SIP: domain='{}', user='{}'".format(sipDomain, self.sipUserName))
                else:
                    self.logger.error("SIP is enabled but some parameters are missing!")

            self.device.loadMixers()
            self._playingNotification = False
            self.notifyRelay1 = False
            self.notifyRelay2 = False
            self.notifyRelay3 = False
            self.notifyRelay4 = False
            if uciGet(uci, "epic", "io", "notify_relay1") == "true":
                self.notifyRelay1 = True
            if uciGet(uci, "epic", "io", "notify_relay2") == "true":
                self.notifyRelay2 = True
            if uciGet(uci, "epic", "io", "notify_relay3") == "true":
                self.notifyRelay3 = True
            if uciGet(uci, "epic", "io", "notify_relay4") == "true":
                self.notifyRelay4 = True

            logMsg = "Initialization is complete. Entering main loop..."
            self.logger.info(logMsg)
            syslog.syslog(logMsg)
            self.runMainLoop()
        except Exception as ex:
            self.logger.error("main loop finished with exception:", exc_info=ex)
            ret_val=1

        self.logger.info("shutting down device...")
        print("shutting down MS700 device...")
        self.logger.info("shutting down serial gw...")
        self.device.shutdown()
        if self.playbackManager:
            print("shutting down playback manager...")
            self.logger.info("shutting down playback manager...")
            self.playbackManager.shutdown()
        if self.tcpApi:
            print("shutting down TCP api...")
            self.logger.info("shutting down TCP api...")
            self.tcpApi.shutdown()

        print("Exiting")
        sys.exit(ret_val)

    def buildDAE(self) -> str:
        # Temperature
        with open("/sys/class/thermal/thermal_zone0/temp") as f:
            temp = int(f.readlines()[0].strip())
            temp = int(temp/1000)

        # Up-time
        with open("/proc/uptime") as f:
            tmp=f.readlines()[0].strip()
            tmp=tmp.split(" ")
            uptime = int(float(tmp[0]))
            utDays = int(uptime/86400)
            utHours= int((uptime-utDays*86400)/3600)
            utMinutes=int((uptime-utDays*86400-utHours*3600)/60)
            uptime = "{:02d}{:02d}{:02d}".format(utDays, utHours, utMinutes)

        # Memory used
        with open("/proc/meminfo") as f:
            tmp=f.readlines()[1].strip()
            tmp=tmp.split(":")[1].strip()
            tmp=tmp.split(" ")[0]
            mem = int(tmp)
            mem = int(mem/1024)

        # Ethernet link down counter
        with open("/sys/class/net/eth0/carrier_changes") as f:
            lnk_chng = int(f.readlines()[0].strip()) - 2

        # PoE type
        if self.device.poeType == 2:
            PoE_type = "at"
        else:
            PoE_type = "no-at"

        # Power consumption
        pwavg = self.device.consumptionAvg1m()
        pwmax = self.device.consumptionMax1m()
        pwmin = self.device.consumptionMin1m()

        message="$DAE:{}:{}:{}:{}:{}:{}:{}:{}\r\n"
        if temp>90:
            message="!!!!!!!!!!!!!!!$DAE:{}:{}:{}:{}:{}:{}:{}:{}!!!!!!!!!!!!!!!\r\n"
        elif temp>80:
            message=">>>>>>>>>>>>>>>$DAE:{}:{}:{}:{}:{}:{}:{}:{}<<<<<<<<<<<<<<<\r\n"
        elif temp>75:
            message="###############$DAE:{}:{}:{}:{}:{}:{}:{}:{}###############\r\n"
        message=message.format(temp, uptime, mem, lnk_chng, PoE_type, pwavg, pwmax, pwmin)
        return message

    def runMainLoop(self):
        loopCount = -20     # We want the first display update to happen much sooner
        lastSipState = None
        btStatus=0x0F
        while not self.stopApplication:
            statusChanged = False

            # Time for refreshing the display
            if loopCount % 20 == 0:    # 1min
                self.device.updateDisplay()

            # Check if device got a proper IP address - other than 127.0.0.1 (loopback interface)
            if self.device.deviceInfo['IP']=="127.0.0.1" and loopCount % 50 == 0:
                self.device.findIp()

            # iterate fw update checks every ~30 seconds
            if self.device.deviceInfo['IP']!="127.0.0.1" and self.fwUpdater is not None and loopCount % 300 == 0:
                self.fwUpdater.checkFwUpdates()

            evt = self.device.readInEvent(1)
            if evt == InOut.PRESSED and (btStatus & 0x01) > 0:
                btStatus &= 0x0E
                logMsg          = "INPUT1 --> ON"
                statusChanged   = True
                self.logger.info(logMsg)
                syslog.syslog(logMsg)

                if self.sipSrc is not None:
                    # if SIP is IDLE, dial self.autodial1 (if not None)
                    # answer if RINGING
                    # hangup if a call is active
                    currentSipState = self.sipSrc.reportSipCallState()
                    if currentSipState == SipCallState.IDLE and len(self.autodial1)>0:
                        logMsg = "SIP AUTODIAL 1: calling '{}'".format(self.autodial1)
                        self.logger.info(logMsg)
                        syslog.syslog(logMsg)
                        self.sipSrc.makeCall(self.autodial1)

                    if currentSipState == SipCallState.RINGING:
                        logMsg = "SIP RINGING + INPUT1 --> ANSWER"
                        self.logger.info(logMsg)
                        syslog.syslog(logMsg)
                        self.sipSrc.answerCall()

                    if currentSipState != SipCallState.IDLE and currentSipState != SipCallState.RINGING:
                        logMsg = "SIP is not IDLE and not receiving a call + INPUT1 --> HANGUP"
                        self.logger.info(logMsg)
                        syslog.syslog(logMsg)
                        self.sipSrc.hangupCall()
            elif evt == InOut.RELEASED:
                logMsg          = "INPUT1 --> OFF"
                btStatus |= 0x01
                statusChanged   = True
                self.logger.info(logMsg)
                syslog.syslog(logMsg)

            evt = self.device.readInEvent(2)
            if evt == InOut.PRESSED and (btStatus & 0x02) > 0:
                btStatus &= 0x0D
                logMsg          = "INPUT2 --> ON"
                statusChanged   = True
                self.logger.info(logMsg)
                syslog.syslog(logMsg)
                # if SIP is IDLE, dial self.autodial2 (if not None)
                # NB: does not anwer if ringing
                if self.sipSrc is not None:
                    currentSipState = self.sipSrc.reportSipCallState()
                    if currentSipState == SipCallState.IDLE and len(self.autodial2)>0:
                        logMsg = "SIP AUTODIAL 2: calling '{}'".format(self.autodial2)
                        self.logger.info(logMsg)
                        syslog.syslog(logMsg)
                        self.sipSrc.makeCall(self.autodial2)
            elif evt == InOut.RELEASED:
                logMsg          = "INPUT2 --> OFF"
                btStatus |= 0x02
                statusChanged   = True
                self.logger.info(logMsg)
                syslog.syslog(logMsg)

            evt = self.device.readInEvent(3)
            if evt == InOut.PRESSED and (btStatus & 0x04) > 0:
                btStatus &= 0x0B
                logMsg          = "INPUT3 --> ON"
                statusChanged   = True
                self.logger.info(logMsg)
                syslog.syslog(logMsg)
                # if SIP is IDLE, dial self.autodial3 (if not None)
                # NB: does not anwer if ringing
                if self.sipSrc is not None:
                    currentSipState = self.sipSrc.reportSipCallState()
                    if currentSipState == SipCallState.IDLE and len(self.autodial3)>0:
                        logMsg = "SIP AUTODIAL 3: calling '{}'".format(self.autodial3)
                        self.logger.info(logMsg)
                        syslog.syslog(logMsg)
                        self.sipSrc.makeCall(self.autodial3)
            elif evt == InOut.RELEASED:
                logMsg          = "INPUT3 --> OFF"
                btStatus |= 0x04
                statusChanged   = True
                self.logger.info(logMsg)
                syslog.syslog(logMsg)

            evt = self.device.readInEvent(4)
            if evt == InOut.PRESSED and (btStatus & 0x08) > 0:
                btStatus &= 0x07
                logMsg          = "INPUT4 --> ON"
                statusChanged   = True
                self.logger.info(logMsg)
                syslog.syslog(logMsg)
                # if SIP is IDLE, dial self.autodial4 (if not None)
                # NB: does not anwer if ringing
                if self.sipSrc is not None:
                    currentSipState = self.sipSrc.reportSipCallState()
                    if currentSipState == SipCallState.IDLE and len(self.autodial4)>0:
                        logMsg = "SIP AUTODIAL 4: calling '{}'".format(self.autodial4)
                        self.logger.info(logMsg)
                        syslog.syslog(logMsg)
                        self.sipSrc.makeCall(self.autodial4)
            elif evt == InOut.RELEASED:
                logMsg          = "INPUT4 --> OFF"
                btStatus |= 0x08
                statusChanged   = True
                self.logger.info(logMsg)
                syslog.syslog(logMsg)

            if statusChanged:
                self._buttonStatus=btStatus

            notificationState = self.playbackManager.hpnIsActive() or self.playbackManager.sipIsActive()
            nwkAudioActive = notificationState or self.playbackManager.bgmIsActive()
            if self._playingNotification != notificationState or self._nwkAudioActive != nwkAudioActive:
                if notificationState:
                    stateStr = "ON"
                    logMsg = "Notify DSP, RED LED --> ON"
                    self.logger.info(logMsg)
                    syslog.syslog(logMsg)
                    self.device.setExclusiveInput(DeviceMs700.LINEIN_STREAMS())
                else:
                    stateStr = "OFF"
                    logMsg = "Notify DSP, RED LED --> OFF"
                    self.logger.info(logMsg)
                    syslog.syslog(logMsg)
                    self.device.setExclusiveInput(0)

                self.device.setRedLed(notificationState)
                if self.notifyRelay1:
                    logMsg = "RELAY1 --> {}".format(stateStr)
                    self.logger.info(logMsg)
                    syslog.syslog(logMsg)
                    self.device.setRelay(1, notificationState)

                if self.notifyRelay2:
                    logMsg = "RELAY2 --> {}".format(stateStr)
                    self.logger.info(logMsg)
                    syslog.syslog(logMsg)
                    self.device.setRelay(2, notificationState)

                if self.notifyRelay3:
                    logMsg = "RELAY3 --> {}".format(stateStr)
                    self.logger.info(logMsg)
                    syslog.syslog(logMsg)
                    self.device.setRelay(3, notificationState)

                if self.notifyRelay4:
                    logMsg = "RELAY4 --> {}".format(stateStr)
                    self.logger.info(logMsg)
                    syslog.syslog(logMsg)
                    self.device.setRelay(4, notificationState)

                if nwkAudioActive:
                    logMsg = "set exclusive input: STREAMS"
                    self.logger.info(logMsg)
                    syslog.syslog(logMsg)
                    self.device.setExclusiveInput(DeviceMs700.LINEIN_STREAMS())
                else:
                    logMsg = "set exclusive input: None"
                    self.logger.info(logMsg)
                    syslog.syslog(logMsg)
                    self.device.setExclusiveInput(0)

            self._playingNotification = notificationState
            self._nwkAudioActive = nwkAudioActive

            if self.sipSrc is not None:
                newSipState = self.sipSrc.reportStatus()
                sipCallState = newSipState["callState"]
                if lastSipState is None or sipCallState != lastSipState["callState"]:
                    stateStr = "????"
                    if sipCallState == SipCallState.IDLE:
                        stateStr = "IDLE"
                    elif sipCallState == SipCallState.CALLING:
                        stateStr = "CALLING dest='{}'".format(newSipState["currentCall"])
                    elif sipCallState == SipCallState.INCOMMING_CALL:
                        stateStr = "INCOMMING"
                    elif sipCallState == SipCallState.RINGING or sipCallState == SipCallState.CALLING_RINGING:
                        stateStr = "RINGING peer='{}'".format(newSipState["currentCall"])
                    elif sipCallState == SipCallState.CONNECTED:
                        stateStr = "CONNECTED peer='{}'".format(newSipState["currentCall"])
                    elif sipCallState == SipCallState.GETTING_NOTIF:
                        stateStr = "NOTIFICATION"
                    logMsg = "SIP Call State --> {}".format(stateStr)
                    self.logger.info(logMsg)
                    syslog.syslog(logMsg)

                registered = newSipState["registered"]
                serverRespCode = newSipState["sipResponseCode"]
                if lastSipState is None or registered != lastSipState["registered"] or serverRespCode != lastSipState[
                    "sipResponseCode"]:
                    logMsg = "SIP Registration: {}, {}".format(registered.upper(), serverRespCode)
                    self.logger.info(logMsg)
                    syslog.syslog(logMsg)
                lastSipState = newSipState

            if statusChanged:
                change = self.getInputChange()
                if change:
                    self.logger.info("broadcasting status update: {}".format(change))
                    self.tcpApi.broadcastStatusUpdate(change)

            loopCount += 1

            # Send the heartbeat to GD
            if self.device.started and time.monotonic() - self._prev_hb_time >= 1.0:
#            if loopCount % 10 == 0:
                self.device.sendMcuCmd("#IIA")
                self._prev_hb_time = time.monotonic()
                if self.device._tty._updateDisplay:
                    self.device.refreshIp()
                    self.device._tty._updateDisplay=False
                    self.device.qccResponse()
                    subprocess.Popen(["barix-wd --pid-file=/var/run/msx00-app.pid --restart"])

            time.sleep(0.1)

        self.logger.info("shutting down MS-x00...")
        #        if self.sipSrc is not None:
        #            self.logger.info("deleting SIP account...")
        #            self.sipSrc.disconnect()

        self.logger.info("tearing down device...")
        self.device.shutdown()

        #        if self.serialGw is not None:
        #            self.logger.info("shutting down serial gw...")
        #            self.serialGw.shutdown()

        self.logger.info("shutting down playback manager...")
        self.playbackManager.shutdown()
        self.logger.info("shutting down TCP api...")
        self.tcpApi.shutdown()
        self.logger.info("leaving main loop")

    def decodeDCommand(self, line:str) -> Tuple[str,str]:
        if len(line)<2:
            return ("",None)

        if line=="DS":
            return ("DS",None)
        elif line=="DA":
            return ("DA",None)
        else:
            # Dsip://id@server
            # DHext
            # DHsip://id@server
            # Dext
            if line.find("DH")==0:
                if len(line)>3:
                    return ("DH",line[2:])
                else:
                    # DH without arguments makes no sense!!
                    return ("",None)
            else:
                return ( "D" , line[1:])

    def executeAbcl(self, line: str) -> str:
        if len(line) == 0:
            return ""

        if line == "#SBL:1":
            if self.amp_mute_status!=1:
                self.device.muteAudioOutputs()
                self.amp_mute_status=1
            return ""
        elif line == "#SBL:0":
            if self.amp_mute_status!=0:
                self.device.unmuteAudioOutputs()
                self.amp_mute_status=0
            return ""
        elif line == "DAX":
            try:
                s = subprocess.check_output(["i2cget", "-y", "0", "0x42", "0x00"])
                s = subprocess.check_output(["i2cdump", "-y", "-r", "0x00-0x1f", "-a", "0", "0x42", "b"])
                s = s.decode("utf-8")
                lines = s.split('\n')
                line = lines[1][4:51] + lines[2][4:51]
                line = line.replace(' ', '')
                result = 'V1,' + line
                return result
            except:
                return "DAX_NOT_SUPPORTED"

        self.logger.info("ABCL command: cmd='{}'".format(line))

        cmd0 = line[0]
        if cmd0 == "T":
            # all other arguments are ignored. might validate empty params...
            reply   = dict()
            m       = self.device.getModel()
            if m==Device.MODEL_MS700:
                reply["m"]  = "MS-700"
            elif m==Device.MODEL_MS500A:
                reply["m"]  = "MA-500A"
            elif m==Device.MODEL_MS300B:
                reply["m"]  = "MS-300B"
            else:
                reply["m"]  = m
            reply["s"] = "IPAM-400"
            reply["v"] = self.sysVersion
            reply["a"] = getMACaddr()
            return json.dumps(reply, indent=4)
        elif cmd0 == "D":
            (cmd, arg) = self.decodeDCommand(line)
            sipState = self.sipSrc.reportSipCallState()
            if cmd == "D":
                # dial SIP <arg> if SIP is IDLE
                if sipState == SipCallState.IDLE and self.sipSrc is not None:
                    logMsg = "ABCL Dial (D) --> dial '{}'".format(arg)
                    self.logger.info(logMsg)
                    syslog.syslog(logMsg)
                    self.sipSrc.makeCall(arg)
            elif cmd == "DS":
                # Hangup any currently active SIP call
                if self.sipSrc is not None:
                    logMsg = "ABCL Hangup (DS) --> hangup SIP call (if any)"
                    self.logger.info(logMsg)
                    syslog.syslog(logMsg)
                    self.sipSrc.hangupCall()
            elif cmd == "DA":
                # answer the incoming SIP call
                if self.sipSrc is not None:
                    logMsg = "ABCL Answer (DA) --> Answer SIP call (if any)"
                    self.logger.info(logMsg)
                    syslog.syslog(logMsg)
                    self.sipSrc.answerCall()
            elif cmd == "DH":
                # dial SIP <arg>
                # if a call is active, terminate it and dial the new one
                if self.sipSrc is not None:
                    if sipState != SipCallState.IDLE:
                        logMsg = "ABCL Dial-Prio (DH) --> hangup current call"
                        self.logger.info(logMsg)
                        syslog.syslog(logMsg)
                        self.sipSrc.hangupCall()
                    logMsg = "ABCL Dial-Prio (DH) --> dial '{}'".format(arg)
                    self.logger.info(logMsg)
                    syslog.syslog(logMsg)
                    self.sipSrc.makeCall(arg)
            else:
                self.logger.warning("unrecognized ABCL D*** command: '{}'".format(line))
        elif line.find("VS=") == 0:
            if len(line) < 4:
                self.logger.warning("malformed VS command: '{}'".format(line))
            else:
                vol = int(line[3:])
                if vol < -79 or vol > 16:
                    self.logger.warning("bad volume in VS command: '{}' (expected=-79..16)".format(vol))
                    return ""
                self.mainVolume = vol
                logMsg = "ABCL Set Volume (VS) --> volume={}".format(self.mainVolume)
                self.logger.info(logMsg)
                syslog.syslog(logMsg)
                uci = Uci()
                uci.set("ae_ms", "volume", "out0", str(self.mainVolume))
                uci.commit("ae_ms")
                self.device.setMainVolume(self.mainVolume)
        elif line == "VQ":
            return str(self.mainVolume)
        elif line == "Q":
            msg = self.getDeviceStatus()
            msg += "\r\n"
            msg += self.buildDAE()
            return msg
        elif line.find("R1=")==0 or line.find("R2=")==0 or line.find("R3=")==0 or line.find("R4=")==0:
            if len(line) < 4:
                self.logger.warning("malformed R1/R2 command: '{}'".format(line))
                return ""
            try:
                relayNum    = int(line[1])
                intVal      = int(line[3])
                relayVal    = True if intVal==1 else False
                self.device.setRelay(relayNum, relayVal)
            except Exception as ex:
                self.logger.error("Something wrong with ABCL '{}'".format(line), exc_info=ex)
        else:
            self.logger.warning("unrecognized ABCL command: '{}'".format(line))

        return ""

    def getInputChange(self) -> str:
#        print('{} {}'.format(self._buttonPStatus, self._buttonStatus))
        dif=self._buttonStatus ^ self._buttonPStatus

        if dif==0:
            return

        self._buttonPStatus=self._buttonStatus

        st=~self._buttonStatus
        curr=1
        forBC=''
        for n in range(1, 5):
#            print('{} {}'.format(dif, curr))
            if dif & curr:
                change=st & curr
                if change > 0:
                    change=1
                if forBC:
                    forBC=forBC+'\r\n'
                forBC=forBC+'statechange,20{},{}'.format(n, change)

            curr=curr << 1

        return forBC

    def getDeviceStatus(self) -> str:
        """
        * "SIP_ID": sip user id
        * STT:
            - S0 --> IDLE
            - S1 --> CALLING
            - S2 --> INCOMING CALL / RINGING
            - S3 --> CONNECTED
            - S4 --> GETTING NOTIFICATION??????
        """
        sipUid = "0"  # invalid SIP ID
        sipState = "S0"

        if self.sipSrc is not None:
            sipUid = self.sipUserName
            state = self.sipSrc.reportSipCallState()
            if state == SipCallState.IDLE:
                sipState = "S0"
            elif state == SipCallState.CALLING:
                sipState = "S1"
            elif state == SipCallState.INCOMMING_CALL or state == SipCallState.RINGING or state == SipCallState.CALLING_RINGING:
                sipState = "S2"
            elif state == SipCallState.CONNECTED:
                sipState = "S3"
            elif state == SipCallState.GETTING_NOTIF:
                sipState = "S4"

        ii = 0x0
        if self.device.readInValue(1) != 0:
            ii |= 0x1
        if self.device.readInValue(2) != 0:
            ii |= 0x2
        if self.device.readInValue(3) != 0:
            ii |= 0x4
        if self.device.readInValue(4) != 0:
            ii |= 0x8

        oo = 0
        if self.device.getOutState(1)==1:
            oo |= 0x1
        if self.device.getOutState(2)==1:
            oo |= 0x2
        if self.device.getOutState(3)==1:
            oo |= 0x4
        if self.device.getOutState(4)==1:
            oo |= 0x8
#        return "{},{},{:02X},{:02X}".format(sipUid, sipState, ii, oo)

        return "{},{},{:02X},{:02X}".format(sipUid, sipState, self._buttonStatus, oo)

    def getAppStatus(self):
        audioStatus = "IDLE"

        sipStatus = "NOT CONFIGURED"
        sipRegStatus = None
        sipRegistrationCode = None
        sipPeer = None
        if self.sipSrc is not None:
            currSipStatus = self.sipSrc.reportStatus()
            sipRegStatus = str(currSipStatus["registered"]).upper()
            sipStatus = self.sipDict[currSipStatus["callState"]]
            sipRegistrationCode = currSipStatus["sipResponseCode"]
            sipPeer = currSipStatus["currentCall"]
            if self.playbackManager.sipIsActive():
                audioStatus = "SIP"

        hpnStatus = "NOT CONFIGURED"
        if self.playbackManager.hpnSource is not None:
            if self.playbackManager.hpnIsActive():
                hpnStatus = "ALIVE"
                audioStatus = "HPN"
            else:
                (state, alive) = self.playbackManager.hpnSource.getState()
                if alive:
                    hpnStatus = "ALIVE"
                else:
                    hpnStatus = "IDLE"

        bgmStatus = "NOT CONFIGURED"
        if self.playbackManager.bgmSource is not None:
            if self.playbackManager.bgmIsActive():
                bgmStatus = "ALIVE"
                audioStatus = "BGM"
            else:
                (state, alive) = self.playbackManager.bgmSource.getState()
                if alive:
                    bgmStatus = "ALIVE"
                else:
                    bgmStatus = "IDLE"

        ioStatus = dict()
        ioStatus["in1"] = "ON" if self.device.readInValue(1)==0 else "OFF"
        ioStatus["in2"] = "ON" if self.device.readInValue(2)==0 else "OFF"
        ioStatus["in3"] = "ON" if self.device.readInValue(3)==0 else "OFF"
        ioStatus["in4"] = "ON" if self.device.readInValue(4)==0 else "OFF"

        ioStatus["out1"] = "ON" if self.device.getOutState(1)==1 else "OFF"
        ioStatus["out2"] = "ON" if self.device.getOutState(2)==1 else "OFF"
        ioStatus["out3"] = "ON" if self.device.getOutState(3)==1 else "OFF"
        ioStatus["out4"] = "ON" if self.device.getOutState(4)==1 else "OFF"

        pwStatus = dict()
        pwStatus["realTime"] = self.device.consumptionRT()
        pwStatus["powerAvg"] = self.device.consumptionAvg1h()
        pwStatus["powerMax"] = self.device.consumptionMax1h()

        status = {
            "sipStatus": "{}".format(sipStatus),
            "sipRegStatus": "{}".format(sipRegStatus),
            "sipRegistrationCode": "{}".format(sipRegistrationCode),
            "sipPeer": "{}".format(sipPeer),
            "hpnStatus": "{}".format(hpnStatus),
            "bgmStatus": "{}".format(bgmStatus),
            "audioStatus": "{}".format(audioStatus),
            "ioStatus": ioStatus,
            "powerStatus": pwStatus
        }
        status = json.dumps(status).encode('utf-8')
        return status

    def identifyXd(self):
        self.device.identifyXd()
