#!/bin/sh

FW_PATH="/etc"
FW_TAR=$FW_PATH/gd-firmware.tar
FW_FILE=*.dfu
TTY="/dev/ttyACM0"
SAFE_PATH="/mnt/data/GD"
LOG_FILE="$SAFE_PATH/gd_upgrade.log"
APP_HOOK=/barix/hooks/application/ae-ms700

log() {
        stamp=$(date '+%Y-%m-%d_%H:%M:%S')
        record="$stamp $1"
        echo $record >> $LOG_FILE
        echo $1
}

mkdir -p $SAFE_PATH

# Get the existing .dfu filename
dfu=""
OLD_VER="0.0.1"
for f in $SAFE_PATH/*.dfu; do
    # Ignore very old fw backup
    if [ "$f" == "/mnt/data/GD/previous-firmware.dfu" ]; then continue; fi

    if [ -f $f ]; then
        dfu=$f
        OLD_VER=$(basename $dfu .dfu)
        OLD_VER=${OLD_VER##*v}
        OLD_VER=$(echo "$OLD_VER" | sed 's/_/\./g')
        break
    fi
done
echo "Existing .dfu filename: $dfu"

# If there is a FW update archive available, extract it to a safe place
# and update the .dfu filename
tmp_dfu=""
ARCH_VER="0.0.1"
if [ -f $FW_TAR ]; then
    # Get rid of very old fw backup
    rm -f /mnt/data/GD/previous-firmware.dfu || true                      
                
    log "GD firmware upgrade file found"
    tmp_dfu=$(tar -tf $FW_TAR)

    # Extract the FW from the archive
    tar -xf $FW_TAR -C $FW_PATH
    rm -f $FW_TAR

    # Extract new FW version
    ARCH_VER=$(basename $tmp_dfu .dfu)
    ARCH_VER=${ARCH_VER##*v}
    ARCH_VER=$(echo "$ARCH_VER" | sed 's/_/\./g')

    COMP=$(echo -e "$OLD_VER\n$ARCH_VER\n" | sort -V -c 2>&1)
    COMP=${COMP: -1}
    if [ "$COMP" == "2" ]; then
        # Move dfu file to safe place
        log "Update $OLD_VER will be replaced by $ARCH_VER"
        rm -f $SAFE_PATH/*.dfu
        mv "$FW_PATH/$tmp_dfu" $SAFE_PATH
        dfu="$tmp_dfu"
        echo "Update .dfu filename: $dfu"
        dfu=$SAFE_PATH/$dfu
    else
        # Remove the extracted .dfu file
        rm -f $FW_PATH/*.dfu
    fi
fi

# Check if GD already in dfu mode
in_dfu=$(dfu-util -l)
if [[ ${in_dfu} = *"Found DFU"* ]]; then
    log "GD found in DFU mode."

    if [ ! -z "$dfu" ]; then
        log "Uploading the fw ($dfu)"

        # Upload the firmware
        r=$(dfu-util -d ,baae:df11 -a 0 -s 0x0800E000:leave -D $dfu)
        if [[ ${r} == *"successfully"* ]]; then
            log "Success!"
            log "Rebooting."
        else
            log "Update failed. Rebooting."
        fi

        reboot
        exit 0
    fi
fi

# If no update archive available exit
if [ "$ARCH_VER" == "0.0.1" ]; then
    echo "No update archive found. Exiting."
    exit 0
fi

# Stop any running MS700 applications
log "Stopping main application, if it's running."
$APP_HOOK stop
sleep 5

# Get installed FW version
CURR_VER=$(python3 gdv.py)
echo "Current version: '$CURR_VER'"
echo "Upgrade version: '$ARCH_VER'"

if [ "$CURR_VER" == "$ARCH_VER" ]; then
    log "The new and old versions are the same. No upgrade performed."
    rm -f "$FW_PATH/$tmp_dfu"
    $APP_HOOK start
    exit 0
fi

COMP=$(echo -e "$CURR_VER\n$ARCH_VER\n" | sort -V -c 2>&1)
COMP=${COMP: -1}
if [ "$COMP" != "2" ]; then
    log "The new version is older than the installed one. No upgrade performed."
    rm -f "$FW_PATH/$tmp_dfu"
    $APP_HOOK start
    exit 0
fi


for arch in $SAFE_PATH/*.dfu; do
    if [ -f $arch ]; then
        echo "" > $LOG_FILE
        log "Starting normal update"

        ### Perform the update
        log "Upgrading from version $CURR_VER to $ARCH_VER"

        # Check if GD already in dfu mode
        tst=$(dfu-util -l)
        if [[ ${tst} != *"Found DFU"* ]]; then
            # Wait for GD to get ready
            gd_ready=1
            count=0
            read -r reply < $TTY
            while [ -z $reply ]; do
                gd_ready=0
                read -r reply < $TTY
                if [ ! -z $reply ];then
                    gd_ready=1
                    log "GD is ready ($count)."
                    break
                fi

                if [ "$count" = "12" ]; then
                    break
                fi

                count=$((count+1))
                sleep 5
                log "$count"
            done

            if [ "$gd_ready" = "0" ]; then
                log "GD never got ready."
                rm -f $FW_PATH/*.dfu
                break
            fi
            log "GD is ready."

            # Switch GD to DFU mode
            echo -ne "\r\n" > $TTY
            echo -ne "#Z96\r\n" > $TTY
            sleep 5

            # Check if the switching succeeded
            count=0
            while true; do
                tst=$(dfu-util -l)
                if [[ ${tst} = *"Found DFU"* ]]; then break; fi

                count=$((count++))
                if ((count>12)); then break; fi
                sleep 5
            done
            if ((count>=12)); then
                log "Couldn't switch to DFU mode. Exiting."
                rm -f $FW_PATH/*.dfu
                break
            fi
            log "Switched to dfu mode"
        fi

        # Upload the firmware
        r=$(dfu-util -d ,baae:df11 -a 0 -s 0x0800E000:leave -D $arch)
        if [[ ${r} == *"successfully"* ]]; then
            log "Success!"
        else
            log "Update failed. Exiting dfu mode."
            dfu-util -d ,baae:df11 -a 0 -s 0x08010000:leave -D /dev/null
        fi

        # Clean-up
        log "Rebooting."
        reboot

        break
    else
        echo "No FW archive found."
        exit 0
    fi
done

$APP_HOOK start
