
import logging
import threading
import time
import serial
import os, fnmatch
from typing import *

from time import sleep
from threading import Timer

NOT_LOGGED_CMDS = ['#IIA', '#SDC', '#SDT', '#SBL']

class IoController:
    banzai = False

    def __init__(self,port:str):
        self._logger    = logging.getLogger("gd")
        self._logger.setLevel(logging.INFO)
        self._lock      = threading.Lock()

        self._serialPortErrorCount=0
        self._updateDisplay=False
        
        if not os.path.exists(port):
            port='/dev/'+fnmatch.filter(os.listdir('/dev'), 'ttyACM*')[0]
        self._logger.info("Serial port: {}".format(port))
        print("Serial port: {}".format(port))
        
        self._stop          = False
        self._tty           = serial.Serial()
        self._tty.port      = port
        self._tty.baudrate  = 115200
        self._tty.bytesize  = serial.EIGHTBITS
        self._tty.parity    = serial.PARITY_NONE
        self._tty.stopbits  = serial.STOPBITS_ONE
        self._tty.xonxoff   = False
        self._tty.rtscts    = False
        self._tty.dsrdtr    = False
        self._tty.rs485_mode= None
        self._tty.timeout   = 0.025
        self._tty.open()
        print("Serial port opened")
        
        # TX queue, one per port...
#        self._txQueue       = dict()
#        for i in range(1,6):
#            self._txQueue[i]= list()

        self._rdBuf         = ""
        self._ttyRxMsgQueue    = list()
        self._ttyRxThread           = threading.Thread(target=self._ttyEventLoop)
        self._ttyRxThread.start()

    def _pRequireLock(self) -> None:
        if not self._lock.locked():
            raise Exception("Lock needs to be aquired!")

    def readMessages(self) -> List[Tuple[int, bytearray]]:
        """
        Read all queued incoming messages
        Returns a list where each entry is a tuple with first element the
        port number of the message and the second element the actual message,
        as a byte array, for support of non-ascii payloads
        """
        with self._lock:
            if len(self._ttyRxMsgQueue) > 0:
                msgs = list()
                rawMsgs = self._ttyRxMsgQueue.copy()
                self._ttyRxMsgQueue.clear()
                for entry in rawMsgs:
                    try:
                        if entry.startswith("#CSP:"):
                            tokens = entry.split(":")
                            msgs.append( (int(tokens[1]), bytearray.fromhex(tokens[2])) )
                        else:
                            msgs.append((0, bytearray(entry,'ascii')))
                    except Exception as e:
                        self._logger.error("Error handling raw message: '{}'".format(entry), exc_info=e)
                return msgs
            return list()


    def sendBytes(self, msg:bytes, port: int ):
        cmd = "#SSP:{}:{}".format(port,msg.hex())
        self._logger.info("TX -> {}: '{}'".format(port,msg))
        self.runMcuCmd(cmd)


    def sendMessage(self, msg:str, port: int ):
        cmd = "#SSP:{}:{}".format(port,msg.encode("ascii").hex())
        self._logger.info("TX -> {}: '{}'".format(port,msg))
        self.runMcuCmd(cmd)


    def runMcuCmd(self,cmd:str, prefixes:List[str]=[] ,timeout:int = 200 ) -> Tuple[int,List[str]]:
        """
        Runs an MCU command
        cmd - string with the command (\\r\\n is added automatically). Maximum
             command length is XXX <----- PLEASE COMPLETE HERE!!!!
        timeout - timeout, in milliseconds (default: 200)
        prefixes - list of reply prefixes that will be used to detect
                    non-default replies
        returns a tuple with:
            first element (int): 0 on success, failure codes TBD
            second element (str): the response to the command
        Must be run in a critical region
        """
        
        # Disable HeartBeat timer for 1 sec
        self._noHeartBeat=True
        
        maxLen      = 200
        if len(cmd)>maxLen:
            self._logger.warning("GD COMMAND is too big (maxLen={}): '{}'".format(maxLen,cmd))
            return -1, ""

        with self._lock:
            if cmd[:4] not in NOT_LOGGED_CMDS or (self.banzai and cmd == '#SBL:0'):    # Do not log these commands
                self._logger.info("GD COMMAND TX: {}".format(cmd))
            expires = time.monotonic_ns() + timeout*1000000

            replies  = list()
            try:
                txdata  = (cmd+"\r\n").encode("utf-8")
                self._tty.write(txdata)
                while time.monotonic_ns()<expires:
                    self._logger.debug("reading line...")
                    ttyLines    = self._pReadLinesFromTty()
                    while len(ttyLines)>0:
                        self._serialPortErrorCount=0
                        line    = ttyLines.pop(0)
                        if type(line)=="tuple":
                            (_, tmp)=line
                            tmp=tmp.decode("utf-8")
                            if tmp.find("#CPL")<0:
                                print(tmp)
                        if line=="OK":
                            self._logger.debug("GD COMMAND '{}' --> OK, {} replies".format(cmd,len(replies)))
                            if len(ttyLines)>0:
                                self._logger.debug("saving for later: " + ' '.join(ttyLines) )
                                self._ttyRxMsgQueue.extend(ttyLines)
                            return (0, replies)

                        if line.startswith("ERR"):
                            self._logger.info("GD COMMAND '{}' --> ERROR".format(cmd))
                            if len(ttyLines)>0:
                                self._logger.info("saving for later: " + ' '.join(ttyLines) )
                                self._ttyRxMsgQueue.extend(ttyLines)
                            self._ttyRxMsgQueue.extend(ttyLines)
                            return (1,replies)

                        saveLine        = True
                        for prefix in prefixes:
                            if line.startswith(prefix):
                                self._logger.debug("GD COMMAND '{}' --> add reply '{}'".format(cmd,line))
                                replies.append(line)
                                saveLine= False
                                break
                        if saveLine:
                            self._logger.debug("saving for later: '{}'".format(line) )
                            self._ttyRxMsgQueue.append(line)
                self._logger.info("GD COMMAND '{}' --> TIMEOUT".format(cmd))
                return (2,"")
            except Exception as e:
                self._logger.error("Exception running GD command: ", exc_info=e )
                self._serialPortErrorCount += 1
                if self._serialPortErrorCount > 3:
                    self._tty.close()
                    self._updateDisplay=True
                    port='/dev/'+fnmatch.filter(os.listdir('/dev'), 'ttyACM*')[0]
                    self._logger.info("Switching serial port to: {}".format(port))
                    print("Switching serial port to: {}".format(port))
                    self._tty.port=port
                    self._tty.open()
                    self._serialPortErrorCount=0
                    time.sleep(5)
                    
                return (-1,None)
    

    def _pReadLinesFromTty(self) -> List[str]:
        """
        Read available lines from the tty port and returns a list of whole lines.
        Any incomplete lines are kept in buffer for the next call
        """
        self._pRequireLock()
        lines   = list()
        while not self._stop:
            try:
                rxBytes     = list(self._tty.read(64))
            except Exception as ex:
                if not self._stop:
                    self._logger.error("Exception reading tty:",exc_info=ex)
                return lines

            if len(rxBytes)==0:
                #print(".", end='', flush=True)
                return lines
            

            for b in rxBytes:
                c   = chr(b)
                self._rdBuf    += c
                if c=='\n':
                    line        = self._rdBuf.strip()
                    #self._logger.debug("TTY RX: '{}'".format(line))
                    lines.append(line)
                    self._rdBuf = ""
        return []


    def _ttyEventLoop(self):
        while not self._stop:
            with self._lock:
                self._ttyRxMsgQueue.extend( self._pReadLinesFromTty() )
            
            time.sleep(0.02)

    def shutdown(self):
        """
        Shutdown operation of the serial GW and disconnect
        from the serial port
        """
        self._stop = True
        self._tty.close()
        self._tty.cancel_read()
        self._logger.info("waiting for worker thread to stop")
        self._ttyRxThread.join()


    def writeSideChannel(self,data:str) -> int :
        """
        Write data to the sidechannel of the serial port
        data - string to write into the serial port, regardless of
               the operation mode.
        """
        with self._lock:
            return self._tty.write(data.encode("utf-8"))
