#!/usr/bin/env python3

import os
import threading
from urllib.parse import urlparse
import time
import signal
import logging
from logging.handlers import RotatingFileHandler
import syslog

from serial.serialposix import Serial

import sys
from uci import Uci
from typing import *
import random
import json
import traceback
import subprocess


from barix.io import leds
from barix.licensing import *
from barix import HttpMethod
import baco

from http_api import HttpApi
from playback_manager import *
from fw_updater import FwUpdater
from tcp_api import TcpApi
from serial_gw import SerialGw
from device_factory import DeviceFactory
from device import Device
from inout import InOut
from epic_app import EpicApp



def uciGet( uci: Uci, config: str, section: str, opt: str) -> str:
    try:
        return uci.get(config,section,opt)
    except:
        return None


def decodeDCommand(line:str) -> Tuple[str,str]:
    if len(line)<2:
        return ("",None)

    if line=="DS":
        return ("DS",None)
    elif line=="DA":
        return ("DA",None)
    else:
        # Dsip://id@server
        # DHext
        # DHsip://id@server
        # Dext
        if line.find("DH")==0:
            if len(line)>3:
                return ("DH",line[2:])
            else:
                # DH without arguments makes no sense!!
                return ("",None)
        else:
            return ( "D" , line[1:])


if __name__ == "__main__":
    HTTP_API_PORT   = 48081
    logger = logging.getLogger("main")
    rotHandler = RotatingFileHandler("/var/log/epic.log", maxBytes=2*1024*1024, backupCount=3)
    logging.basicConfig(level=logging.INFO,handlers=[rotHandler],format='%(asctime)s %(name)s %(levelname)s %(message)s')
    baco.initializeLogger("/var/log/baco.log",baco.LogPriority.INFO,2000000,5,False)

    #
    # early development units have either:
    # * core-image-ae-msx00
    # * ms700-fw
    #
    # production version is built as:
    # * core-image-ae-ms700
    #
    # All of the above should allow operation of the software

    found, _ = findLicense("core-image-ae-msx00")
    if not found:
        found, _ = findLicense("core-image-ae-ms700")
    if not found:
        found, _ = findLicense("ms700-fw")
    if not found:
        found, _ = findLicense("barix-development-platform")

    if not found:
        logger.error("required license is missing. Exit.")
        sys.exit(1)

#    uciSet(Uci(), "ae_ms", "apcore", "enable", "false")
    baco.initializeAudioLib()
    DeviceFactory.createDevice()
    
    # Check PoE switch type
    if not DeviceFactory.instance().isPoE_OK():
        logger.error("PoE: Insufficient power")
    logger.info("PoE: OK")
    
    EpicApp.setup(DeviceFactory.instance())
    app     = EpicApp.get()
    httpApi = HttpApi("127.0.0.1", HTTP_API_PORT, app)
    httpApi.launch()
    app.run()
    logger.info("shutting down http API...")
    httpApi.stop()
    print("ALL FINISHED")
