#!/usr/bin/lua
local logging = require("logging");
--local log_console = require("logging.console");
local log_file = require("logging.rolling_file");

local logger = log_file{
    filename="/var/log/arcanix.log",
    maxFileSize=2048*1024,
    maxBackupIndex=3,
    logLevel=logging.DEBUG,
    logPattern="%date %level %file:%line %message\n"
};

--[[
local logger = log_console{
    timestampPattern="%Y-%m-%d %X",
    logPattern="%date %level %file:%line %message\n"
};
]]

-- always set default logger before load arcanix module
logging.defaultLogger(logger);

local uv = require("luv");
local arcanix = require("arcanix");
local uci_handler = require("arcanix.uci_handler")

local DEVICE_MODEL="MS-700"
local RELOAD_SETTINGS_URL = "http://127.0.0.1:48081/api/v1/reload-settings";
local IDENTIFY_XD_URL = "http://127.0.0.1:48081/api/v1/identify-xd";
local REFRESH_IP = "http://127.0.0.1:48081/api/v1/refresh-ip";

local function finish_app()
    arcanix.stop()
    os.exit(0)
end

local term_signal = uv.new_signal()
term_signal:start("sigint", finish_app)
term_signal:start("sigterm", finish_app)

local http_work = uv.new_work(
    function(url)
        local ll = require("logging");
        local logger = ll.defaultLogger();
        logger:info("sending http request to %s", url);

        http = require("socket.http");
        http.TIMEOUT = 5000;

        local request_body = "";
        local response_body = {};
        local ltn12 = require("ltn12")

        local res, code, response_headers = http.request{
          url = url,
          method = "PUT",
          headers =
            {
                ["Content-Length"] = #request_body;
            },
            source = ltn12.source.string(request_body),
            sink = ltn12.sink.table(response_body),
        }

        return code;
    end,
    function(code)
        logger:debug("Return code %s", code);
    end
);

local function reloadSettings()
    http_work:queue(RELOAD_SETTINGS_URL);
end

local shell_work = uv.new_work(
    function(cmd)
        local ll = require("logging");
        local logger = ll.defaultLogger();
        logger:info("executing shell cmd %s", cmd);
        result = os.execute(cmd);
        reloadSettings();
        return result;
    end,
    function(ret)
        logger:info("Shell returned %s", ret);
    end
);

local function applyNetworkSettings()
    shell_work:queue("/etc/init.d/networking restart");
    uv.sleep(10000);
    http_work:queue(REFRESH_IP);
end

local function restartDSP()
    shell_work:queue("barix-wd --pid-file=/var/run/ms700-dsp.pid --restart");
end

local function rebootDevice()
    shell_work:queue("/sbin/reboot");
end

local function applySNMP()
    shell_work:queue("/barix/apps/epic-sip/applySNMP.sh");
end

local uci_map = {
    ["network.eth0.proto"] =         {uuid=0x00010000, cmd=applyNetworkSettings},
    ["network.eth0.ipaddr"] =        {uuid=0x00010001, cmd=applyNetworkSettings},
    ["network.eth0.netmask"] =       {uuid=0x00010002, cmd=applyNetworkSettings},
    ["network.eth0.gateway"] =       {uuid=0x00010003, cmd=applyNetworkSettings},
    ["network.eth0.dns_type"] =      {uuid=0x00010004, cmd=applyNetworkSettings},
    ["network.eth0.dns1"] =          {uuid=0x00010005, cmd=applyNetworkSettings},
    ["network.eth0.dns2"] =          {uuid=0x00010006, cmd=applyNetworkSettings},
    ["epic.server.ipaddr"] =         {uuid=0x00010007, cmd=reloadSettings},
    ["ae_ms.mixer.ch1"] =            {uuid=0x00010009, cmd=reloadSettings},
    ["ae_ms.mixer.ch2"] =            {uuid=0x0001000A, cmd=reloadSettings},
    ["ae_ms.mixer.ch3"] =            {uuid=0x0001000B, cmd=reloadSettings},
    ["ae_ms.mixer.ch4"] =            {uuid=0x0001000C, cmd=reloadSettings},
    ["ae_ms.mixer.ch5"] =            {uuid=0x0001000D, cmd=reloadSettings},
    ["ae_ms.mixer.ch6"] =            {uuid=0x0001000E, cmd=reloadSettings},
    ["ae_ms.mixer.ch100"] =          {uuid=0x0001000F, cmd=reloadSettings},
    ["ae_ms.volume.out1"] =          {uuid=0x00010010, cmd=reloadSettings},
    ["ae_ms.volume.out2"] =          {uuid=0x00010011, cmd=reloadSettings},
    ["ae_ms.volume.out3"] =          {uuid=0x00010012, cmd=reloadSettings},
    ["ae_ms.volume.out4"] =          {uuid=0x00010013, cmd=reloadSettings},
    ["ae_ms.eq.band1"] =             {uuid=0x00010014, cmd=reloadSettings},
    ["ae_ms.eq.band2"] =             {uuid=0x00010015, cmd=reloadSettings},
    ["ae_ms.eq.band3"] =             {uuid=0x00010016, cmd=reloadSettings},
    ["ae_ms.eq.band4"] =             {uuid=0x00010017, cmd=reloadSettings},
    ["ae_ms.eq.band5"] =             {uuid=0x00010018, cmd=reloadSettings},
    ["ae_ms.eq.enable"] =            {uuid=0x00010019, cmd=reloadSettings},
    ["ae_ms.serial.egress_list_1"] = {uuid=0x0001001A, cmd=reloadSettings},
    ["ae_ms.serial.egress_list_2"] = {uuid=0x0001001B, cmd=reloadSettings},
    ["ae_ms.serial.egress_list_3"] = {uuid=0x0001001C, cmd=reloadSettings},
    ["ae_ms.serial.egress_list_4"] = {uuid=0x0001001D, cmd=reloadSettings},
    ["ae_ms.serial.egress_list_5"] = {uuid=0x0001001E, cmd=reloadSettings},
    ["ae_ms.serial.egress_list_6"] = {uuid=0x0001001F, cmd=reloadSettings},
    ["ae_ms.serial.config_1"] =      {uuid=0x00010020, cmd=reloadSettings},
    ["ae_ms.serial.config_2"] =      {uuid=0x00010021, cmd=reloadSettings},
    ["ae_ms.serial.config_3"] =      {uuid=0x00010022, cmd=reloadSettings},
    ["ae_ms.serial.config_4"] =      {uuid=0x00010023, cmd=reloadSettings},
    ["ae_ms.serial.config_5"] =      {uuid=0x00010024, cmd=reloadSettings},
    ["ae_ms.volume.out0"] =          {uuid=0x00010025, cmd=reloadSettings},
    ["ae_ms.mixer.enable1"] =        {uuid=0x00010026, cmd=reloadSettings},
    ["ae_ms.mixer.enable2"] =        {uuid=0x00010027, cmd=reloadSettings},
    ["ae_ms.mixer.enable3"] =        {uuid=0x00010028, cmd=reloadSettings},
    ["ae_ms.mixer.enable4"] =        {uuid=0x00010029, cmd=reloadSettings},
    ["ae_ms.mixer.enable5"] =        {uuid=0x0001002A, cmd=reloadSettings},
    ["ae_ms.mixer.enable6"] =        {uuid=0x0001002B, cmd=reloadSettings},
    ["ae_ms.mixer.enable10"] =       {uuid=0x0001002C, cmd=reloadSettings},
    ["ae_ms.volume.enable0"] =       {uuid=0x0001002D, cmd=reloadSettings},
    ["ae_ms.volume.enable1"] =       {uuid=0x0001002E, cmd=reloadSettings},
    ["ae_ms.volume.enable2"] =       {uuid=0x0001002F, cmd=reloadSettings},
    ["ae_ms.volume.enable3"] =       {uuid=0x00010030, cmd=reloadSettings},
    ["ae_ms.volume.enable4"] =       {uuid=0x00010031, cmd=reloadSettings},
    ["ae_ms.poe.override"] =         {uuid=0x00010032, cmd=rebootDevice},
    ["ae_ms.banzai.enable"] =        {uuid=0x00010033, cmd=restartDSP},
    ["ae_ms.banzai.threshold"] =     {uuid=0x00010034, cmd=restartDSP},
    ["ae_ms.speaker.impedance"] =    {uuid=0x00010035, cmd=reloadSettings},
    ["epic.snmp.enable"] =           {uuid=0x00010036, cmd=applySNMP}
};

local handler = uci_handler:new()

for uci, param in pairs(uci_map) do
    handler:add_configuration(uci, param.uuid, param.cmd);
end

-- wait for the dongle to be ready
uv.sleep(30000);

arcanix.start(uv, arg[1], handler, function(value)
    if value == "beep-tone" then
        shell_work:queue("aplay -D plug:rtp_dmix beep.wav");
    elseif value == "xd-leds" then
        http_work:queue(IDENTIFY_XD_URL);
    end
end, DEVICE_MODEL);

while true do
    uv.run();
end
