import logging
import os.path
import syslog
import threading
from threading import Timer
import subprocess
import alsaaudio
import time
import socket
from typing import *
from select import select


from uci import Uci
from barix import HttpMethod
from barix.io import leds

from device import Device
from inout import InOut
from serial_gw import SerialGw
from io_controller import IoController
from uci_util import uciGet, uciSet

    
def isValidLineIn( lineId:int ) -> bool:
    return (lineId>0 and lineId<7) or lineId==100


class InputPin:
    def __init__(self,gdPin:int):
        self.events     = list()
        self.gdPin      = gdPin

class OutRelay:
    def __init__(self,gdPin:int):
        self.gdPin      = gdPin
        self.value      = 0


class EpicLineIn:
    def __init__(self):
        self.enabled    = False
        self.volume     = 0


class EpicLineOut:
    def __init__(self):
        self.enabled    = False
        self.volume     = 0


class DeviceMs700(Device):
    started=False

    # speaker consumption taken by measuring the peak value using 
    # speaker-test tool playing a 1KHz tone.
    # Then, the RMS value is taken of the peak and the power consumption
    # is calculated using the formula P = (Vrms^2)/Impedance
    # Peak values for each LineOut Atenuation:
    #  -6dB=10.8V; -7.5dB=10.4V; -9dB=9.8V; -10.5dB=8.2V; -12dB=7V
    #  -13.5dB=5.8V; -15dB=5V
    SPEAKER_CONSUMPTION = {
        "2": {
            0: 29.16,
            -7.5: 27,
            -9: 24,
            -10.5: 16.81,
            -12: 12.00,
            -13.5: 8.4,
            -15: 6.25
        },
        "4": {
            0: 14.58,
            -7.5: 13.52,
            -9: 12,
            -10.5: 8.41,
            -12: 6.12,
            -13.5: 4.20,
            -15: 3.12
        },
        "8": {
            0: 7.2,
            -7.5: 6.76,
            -9: 6,
            -10.5: 4.2,
            -12: 3,
            -13.5: 2.1,
            -15: 1.56
        }
    }

    deviceInfoP = {
        'IP':    "",
        'fwVer': "",
        'gdVer': "",
        'pwRt':  0.0,
        'pwAvg': 0.0,
        'pwMax': 0.0
    }

    deviceInfo = {
        'IP':    "",
        'fwVer': "",
        'gdVer': "",
        'pwRt':  0.0,
        'pwAvg': 0.0,
        'pwMax': 0.0
    }

    poeType=0
    PoEmessageActive = False
    MoEmsgStartTime = None
    
    ppwRT = 0
    pwRT = 0
    
    lastTime1m = time.time()
    pwAcc1m=0
    pwAvg1m=0
    pwMin1m=999
    pwMax1m=0
    count1m=0
    
    lastTime1h = time.time()
    pwAcc1h=0
    pwAvg1h=0
    pwMax1h=0
    count1h=0
    
    @staticmethod
    def LINEIN_STREAMS():
        return 100

    @staticmethod
    def LINEIN_XD():
        return 4

    @staticmethod
    def LINEIN_IDS():
        return [1, 2, 3, 4, 5, 6, 100]

    def __init__(self,model:str) -> None:
        Device.__init__(self,model)
        self._logger    = logging.getLogger("device")
        self._lock      = threading.Lock()
        self._idModeTmr = None
        self._displayTimeout    = 0
        self._muteStatus = None
        
        # Power consumed by the device when it's not playing. Set once on startup.
        self._minPower = 12
        try:
            # file filled by a script that runs before this application
            # value in amps
            with open('/tmp/consumption', 'r') as fp:
                self._minPower = float(fp.read().split('=')[1])
                self._minPower *= 24 # multiply with voltage to get power
                self._logger.info(f'Minimum power consumption: {self._minPower}W')
        except Exception as exc:
            self._logger.error('Error reading power consumption')

        uci             = Uci()
        serialGwMode    = uciGet(uci,"epic","serialgw","mode")
        serialPeerPort  = int(uciGet(uci,"epic","serialgw","port"))
        self._egressList = dict()
        self._mappedSerialPort  = 1

        if serialGwMode!=SerialGw.NOGW and serialGwMode!=SerialGw.SERVER:
            self._logger.error("invalid mode: {}".format(serialGwMode))
            raise Exception("mode is invalid: " + serialGwMode )

        if serialGwMode==SerialGw.SERVER and (serialPeerPort<1 or serialPeerPort>65535):
            self._logger.error("invalid gw port: {}".format(serialPeerPort))
            raise Exception("Invalid GW port: {}".format(serialPeerPort))

        self._gwMode        = serialGwMode
        self._gwPort        = serialPeerPort
        self._stopGw        = False
        self._srvSkt        = None
        self._sktCon        = None
        self._tcpBuf        = bytearray()
        self._tcpRxQueue    = bytearray()
        # /dev/ttyS1 for M400
        self._tty           = IoController('/dev/ttyACM0')
        self._tty.banzai = True if uciGet(uci, "ae_ms", "banzai", "enable") == 'true' else False
        
        self._logger.info("Waiting for GD to come alive...")
        with self._lock:
            while self._tty.runMcuCmd("#SPC:100:0:0:0",[],5000)[0]!=0:
                pass
        self._logger.info("GD is alive!")

        self._lineIn            = dict()
        for i in DeviceMs700.LINEIN_IDS():
            lineIn              = EpicLineIn()
            lineIn.enabled      = True
            lineIn.volume       = -79
            self._lineIn[i]     = lineIn

        self._lineOut           = dict()
        for i in range(0,5):
            lineOut         = EpicLineOut()
            lineOut.enabled = True
            lineOut.volume  = 0
            self._lineOut[i]= lineOut

        self._eqBand        = dict()
        for i in range(1,6):
            self._eqBand[i] = 0
        
        self._inpExcReq     = 0 # exclusive input disabled

        self._ipin1                 = InputPin(4)
        self._ipin2                 = InputPin(3)
        self._ipin3                 = InputPin(2)
        self._ipin4                 = InputPin(1)
        self._opin1                 = OutRelay(3)
        self._opin2                 = OutRelay(4)
        self._opin3                 = OutRelay(1)
        self._opin4                 = OutRelay(2)

        self._redLed    = leds.Led("barix:led1:red")
        self._greenLed  = leds.Led("barix:led1:green")
        self._greenLed.set(False)
        self._redLed.set(False)
        syslog.syslog("RED LED --> OFF")
        syslog.syslog("GREEN LED --> OFF")

        self._workerThread          = threading.Thread(target=self._eventLoop)
        self._workerThread.daemon   = True
        self._workerThread.start()
        
        # Enable display
        execCmd     = "#SDS:1"
        self._tty.runMcuCmd(execCmd)
        
        # FW version
        with open("/barix/info/VERSION", "r") as fp:
            self.deviceInfo['fwVer'] = fp.read()
        time.sleep(1)
        print("FW: {}".format(self.deviceInfo['fwVer']))
        
        # Find device IP
        self.findIp()
        
        self.started=True
        
    def refreshIp(self):
        self.findIp()
        
    def sendMcuCmd(self, c:str):
        self._tty.runMcuCmd(c)
    
    def updateDisplay(self):
        # Update 1st line
        if self.deviceInfo['IP'] != "" and self.deviceInfo['IP'] != self.deviceInfoP['IP']:
            ip = self.deviceInfo['IP'].center(16)
            data = "".join("{:02x}".format(ord(c)) for c in ip).upper()
            execCmd     = "#SDT:1:{}".format(data)
            self._tty.runMcuCmd("#SDC:1:K")
            self._tty.runMcuCmd(execCmd)
            
            self.deviceInfoP['IP'] = self.deviceInfo['IP']
        
        # There is a PoE message displayed. Wait 30sec
#        if self.PoEmessageActive:
#            t = time.time()
#            if t - self.PoEmsgStartTime < 30:
#                return
#            self.PoEmessageActive = False
        
        # There is a PoE message displayed. Do not show the 2nd & 3rd lines
        if self.PoEmessageActive:
            return
        
        # Update 2nd line
        if len(self.deviceInfo['fwVer']) > 2 and len(self.deviceInfo['gdVer']) > 2 and (self.deviceInfo['fwVer'] != self.deviceInfoP['fwVer'] or self.deviceInfo['gdVer'] != self.deviceInfoP['gdVer']):
            tmp1 = "FW" + self.deviceInfo['fwVer']
            if '-' in tmp1:
                tmp1=tmp1.split("-")[0]
            tmp2 = "GD" + self.deviceInfo['gdVer']
            tmp = tmp1 + " " + tmp2
            tmp = tmp.center(16)
            data = "".join("{:02x}".format(ord(c)) for c in tmp).upper()
            execCmd     = "#SDT:2:{}".format(data)
            self._tty.runMcuCmd("#SDC:1:K")
            self._tty.runMcuCmd(execCmd)
            
            self.deviceInfoP['fwVer'] = self.deviceInfo['fwVer']
            self.deviceInfoP['gdVer'] = self.deviceInfo['gdVer']
            
        # Update 3rd line
        if self.pwRT > 0 and self.pwRT != self.ppwRT:
            tmp1 = "{}".format(round(self.pwRT *24, 2))
            tmp2 = "{}".format(self.consumptionAvg1h())
            tmp3 = "{}".format(round(self.pwMax1h * 24, 2))
            tmp = tmp1 + "/" + tmp2 + "/" + tmp3
            
            tmp = tmp.center(16)
            data = "".join("{:02x}".format(ord(c)) for c in tmp).upper()
            execCmd     = "#SDT:3:{}".format(data)
            self._tty.runMcuCmd("#SDC:1:K")
            self._tty.runMcuCmd(execCmd)
            
            self.ppwRT = self.pwRT
        
    def findIp(self):
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.settimeout(0)
        try:
            s.connect(('10.254.254.254', 1))
            IP = s.getsockname()[0]
        except Exception:
            IP = '127.0.0.1'
        finally:
            s.close()
            self.deviceInfo['IP']=IP

    def isPoE_OK(self):
        # Check power type
        PoEfile="/barix/PoE_type.txt"
        PoE_type = 0
        if os.path.exists(PoEfile):
            with open(PoEfile) as f:
                PoE_type = f.readline().strip()
        print("PoE {}".format(PoE_type))
        
        if PoE_type=="type=3":
            self.poeType=3
            command_seq = [
            "#SAL:11",
            "#SDS:1",
            "#SDC:1:R",
            "#SDT:2:506F45206F76657272696464656E",      # 'PoE overridden'
            "#SDT:3:00",
            "#SDC:1:K"
            ]
            
            for cmd in command_seq:
                self._tty.runMcuCmd(cmd)
            
            self.PoEmessageActive = True
            self.PoEmsgStartTime = time.time()
            
            # Draw boxes at the bottom line
#            for i in range(1, 9):
#                execCmd = "#SDI:{}:R".format(i)
#                self._tty.runMcuCmd(execCmd)
            
        elif PoE_type=="type=2":
            self.poeType=2
            command_seq = [
            "#SAL:00",
            "#SDS:1",
            "#SDC:1:K",
            "#SDT:2:00",
            "#SDT:3:00",
            "#SDC:1:K"
            ]
            
            for cmd in command_seq:
                self._tty.runMcuCmd(cmd)
                
            return False
            self._tty.runMcuCmd(execCmd)
        else:
            command_seq = [
            "#SAL:11",
            "#SDS:1",
            "#SDC:1:R",
            "#SDT:2:506F452B207265717569726564",            # 'PoE+ required'
            "#SDT:3:506F776572206C696D69746564203157",      # 'Power limited 1W'
            "#SDC:1:K"
            ]
            
            for cmd in command_seq:
                self._tty.runMcuCmd(cmd)
                
            self.PoEmessageActive = True
            self.PoEmsgStartTime = time.time()

            return False
            
        return True

    def reloadSettings(self) -> None:
        logger      = logging.getLogger("device")
        uci         = Uci()
        for chId in self.LINEIN_IDS():
            val         = uciGet(uci,"ae_ms","mixer","ch{}".format(chId))
            enable      = uciGet(uci,"ae_ms","mixer","enable{}".format(chId))
            gain        = -79
            if len(val)>0:
                try:
                    gain    = int(val)
                    if gain<-79 or gain >16:
                        raise AttributeError("line in {} gain out of bounds: {}".format(chId,gain))
                except:
                    logger.error("cannot read gain for input {}: value='{}'".format(chId,val))
            self._lineIn[chId].enabled  = True if enable=="true" else False
            self._lineIn[chId].volume   = gain

        with self._lock:
            self._pRefreshInputs()

        for chId in range(0,5):
            val         = uciGet(uci,"ae_ms","volume","out{}".format(chId))
            enable      = uciGet(uci,"ae_ms","volume","enable{}".format(chId))
            gain        = -79
            if len(val)>0:
                try:
                    gain    = int(val)
                    if gain<-79 or gain>16:
                        raise AttributeError("line out gain out of range: {}".format(chId,gain))
                except Exception as e:
                    logger.error("cannot set gain for output {}:".format(chId),exc_info=e)
            self._lineOut[chId].volume  = gain
            self._lineOut[chId].enabled = True if enable=="true" else False
            with self._lock:
                self._pRefreshLineOut(chId)

        val     = uciGet(uci, "ae_ms", "talkback", "level_wpa")
        gain    = -79
        if len(val)>0:
            try:
                gain    = int(val)
                if gain<-79 or gain>16:
                    raise AttributeError()
            except:
                self._logger.error("invalid ae_ms.talkback.level_wpa: '{}'".format(val))
        self._tty.runMcuCmd("#SIV:8:{}".format(gain))

        val     = uciGet(uci, "ae_ms", "talkback", "level_rp1")
        gain    = -79
        if len(val)>0:
            try:
                gain    = int(val)
                if gain<-79 or gain>16:
                    raise AttributeError()
            except:
                self._logger.error("invalid ae_ms.talkback.level_rp1: '{}'".format(val))
        self._tty.runMcuCmd("#SIV:7:{}".format(gain))


        for bandId in range(1,6):
            val     = uciGet(uci,"ae_ms","eq","band{}".format(bandId))
            gain    = 0
            if len(val)>0:
                try:
                    gain    = int(val)
                    if gain<-15 or gain>15:
                        raise AttributeError("eq band {} out of range: {}".format(bandId,gain))
                except Exception as e:
                    logger.error("cannog get eq band {}:".format(bandId),exc_info=e)
                self._eqBand[bandId]    = gain

        val = uciGet(uci,"ae_ms","eq","enable")
        self._eqEnable  = True if val=="true" else False
        self._updateEq()

        for ttyId in range(1,7):
            s           = uciGet(uci,"ae_ms","serial","egress_list_{}".format(ttyId))
            if len(s)==0:
                self._egressList[ttyId]  = list()
            else:
                egressList  = s.split(',')
                tmpList     = list()
                goodList    = True
                if s!="0":
                    for p in egressList:
                        if not p.isdecimal():
                            goodList    = False
                            break
                        pval    = int(p)
                        if pval<1 or pval>6 or pval==ttyId:
                            goodList = False
                            break
                        tmpList.append(pval)
                    if not goodList:
                        logger.error("invalid egress list for port {} : '{}'".format(ttyId,s))
                self._egressList[ttyId]  = tmpList if goodList else list()

        for serialId in range (1,6):
            self._logger.info("reloading serial settings for port {}...".format(serialId))
            s       = uciGet(uci,"ae_ms","serial","config_{}".format(serialId))
            if len(s)>0:
                execCmd = "#SSC:{}:{}".format(serialId,s)
                self._tty.runMcuCmd(execCmd)

        arcanixMode = True if uciGet(uci,"ae_ms","arcanix","mode")=="enabled" else False
        arcanixState= os.path.exists("/var/run/ms700-arcanix.pid")
        if arcanixMode!=arcanixState:
            subprocess.Popen(["/barix/hooks/application/ms700-arcanix" , "start" if arcanixMode else "stop"])


    def _getLineOutAttenuation(self, imp):
        if not imp:
            # default value is 2ohm
            imp = "2"

        # The SPEAKER_CONSUMPTION table is not very precise. That's why we need this
        # adjustment.
        if imp == '2':
            powerAdjust = 5
        elif imp == '4':
            powerAdjust = 3
        elif imp == '8':
            powerAdjust = 0
        else:
            self._logger.error(f'Unsupported speaker impedance: {imp}')
            return 0

        for lineoutAtten, spkConsumption in self.SPEAKER_CONSUMPTION[imp].items():
            # total power consumption (speaker + minimum) allowed is 20W
            if (self._minPower + spkConsumption + powerAdjust) < 20:
                self._logger.info(f'minPower {self._minPower}W + spkConsum {spkConsumption}W + adjust {powerAdjust}W < 20W')
                self._logger.info(f"Clip Level adjusted to {lineoutAtten}dB")
                return lineoutAtten

        return 0


    def _getEpicLineIn(self,lineId:int) -> EpicLineIn:
        """
        Retrieve the EPIC line in object
        """
        if lineId not in self.LINEIN_IDS():
            raise Exception("Invalid line in id: {}".format(lineId))

        return self._lineIn[lineId]


    def _getEpicLineOut(self, lineId:int) -> EpicLineOut:
        """
        Retrieve the EPIC line out object
        """
        if lineId<0 or lineId>4:
            raise Exception("Invalid line out id: {}".format(lineId))
        
        return self._lineOut[lineId]


    def _getInputPin(self,id:int) -> InputPin:
        if id==1:
            return self._ipin1
        elif id==2:
            return self._ipin2
        if id==3:
            return self._ipin3
        elif id==4:
            return self._ipin4
        else:
            raise Exception("Invalid pin id: {}".format(id))


    def _getRelay(self,id:int) -> OutRelay:
        if id==1:
            return self._opin1
        elif id==2:
            return self._opin2
        elif id==3:
            return self._opin3
        elif id==4:
            return self._opin4
        else:
            raise Exception("Invalid output relay id: {}".format(id))

    def setRelay(self, relayId:int, stt:bool) -> None:
        opin        = self._getRelay(relayId)
        opin.value  = 1 if stt else 0
        cmd         = "#SOL:{}:{}".format(opin.gdPin,opin.value)
        with self._lock:
            (res,_) = self._tty.runMcuCmd(cmd)
        self._logger.info("RELAY {} --> {}. command result: {}".format(relayId, stt, res))

    def setRedLed(self,stt:bool) -> None:
        self._redLed.set(stt)
        #self._logger.info("SETTING RED LED to {} <---- PLEASE IMPLEMENT!!!".format(stt))

    def setGreenLed(self,stt:bool) -> None:
        self._greenLed.set(stt)
        #self._logger.info("SETTING GREEN LED to {} <---- PLEASE IMPLEMENT!!!".format(stt))

    def readInEvent(self,id:int) -> int:
        """
        Read one input event from input 
        id - id of the input (1 or 2)
        returns - InOut.IDLE, InOut.PRESSED or InOut.RELEASED
        """
        pin = self._getInputPin(id)
        if len(pin.events)==0:
            return InOut.IDLE
        else:
            return pin.events.pop(0)

    def readInValue(self, pinId:int) -> int:
        """
        Read current value of input pin
        id - id of the pin (1 or 2)
        return 0 (low) or 1(high)
        """
        pin = self._getInputPin(pinId)
        cmd = "#QIL:{}".format(pin.gdPin)
        with self._lock:
            (res,replies) = self._tty.runMcuCmd(cmd,["#QIL:"],1000)
            if res==0:
                if len(replies)==0:
                    self._logger.error("PROTOCOL VIOLATION: #QIL success, but no replies returned!")
                    return 0
                comps   = replies[0].split(":")
                return int(comps[2])
            else:
                self._logger.error("Error running command: '{}'".format(cmd))
                return 0

    def getOutState(self, outputId: int ) -> int:
        if outputId == 1:
            return self._opin1.value
        elif outputId==2:
            return self._opin2.value
        elif outputId==3:
            return self._opin3.value
        elif outputId==4:
            return self._opin4.value
        else:
            raise IndexError("Invalid index: {}".format(outputId))
    def getIoMode(self,id:int) -> str:
        """
        Compatibility stub for the MS-700.
        The result can be safely ignores.
        Always returns "INPUT"
        """
        return "INPUT"

    def getSipPcmc(self) -> str:
        """
        Get the name of the SIP PCM capture device (microphone)
        """
        return "plug:analog_in_right"

    def getSipPcmp(self) -> str:
        """
        Get the name of the SIP PCM playback device (audio out)
        """
        return "sip_out_play"

    def getModel(self) -> str:
        """
        Get the device model
        """
        return self._model

    def getRtpBgmCh(self) -> str:
        """
        Get the alsa channel for RTP BGM
        """
        return "rtp_ch_1"

    def getRtpHpnCh(self) -> str:
        """
        Get the alsa channel for RTP HPN
        """
        return "rtp_ch_2"

        
    def loadMixers(self) -> None:
        self._setMixerParam("Line In", "cap")
        self._setMixerParam("ADC Gain", "0dB")
        self._setMixerParam("Mic1", "nocap")
        alsaaudio.PCM(device='plug:rtp_ch_1')
        alsaaudio.PCM(device='plug:rtp_ch_2')
        alsaaudio.PCM(device='plug:rtp_ch_3')
        alsaaudio.PCM(device='plug:rtp_ch_4')
        alsaaudio.PCM(device='plug:rtp_ch_5')
        alsaaudio.PCM(device='plug:sip_out_play')

    def _ivi9(self,msg,ingressPort):
        self._pRequireLock()
        uci = Uci()
        notifMsg = ""
        for intLine in [5, 6]:
            lineIn = self._getEpicLineIn(intLine)
            if lineIn.volume < 16:
                lineIn.volume += 5
                if lineIn.volume > 16:
                    lineIn.volume = 16
            uciSet(uci, "ae_ms", "mixer", "ch{}".format(intLine), str(lineIn.volume))
            notifMsg += "$CVI:{}:{}\r\n".format(intLine, lineIn.volume)
        self._pRefreshInputs()
        self._sendMsg("{}\r\n".format(msg), ingressPort)
        self._bcastMsgExcl(notifMsg)

    def _dvi9(self,msg,ingressPort):
        self._pRequireLock()
        uci = Uci()
        notifMsg = ""
        for intLine in [5, 6]:
            lineIn = self._getEpicLineIn(intLine)
            if lineIn.volume > -79:
                lineIn.volume -= 5
                if lineIn.volume < -79:
                    lineIn.volume = -79
            uciSet(uci, "ae_ms", "mixer", "ch{}".format(intLine), str(lineIn.volume))
            notifMsg += "$CVI:{}:{}\r\n".format(intLine, lineIn.volume)
        self._pRefreshInputs()
        self._sendMsg("{}\r\n".format(msg), ingressPort)
        self._bcastMsgExcl(notifMsg)

    def handleAeCommand(self, msg:str, ingressPort:int) -> bool:
        """
        Handle one command.
        msg - the command to handle
        port - the ingress port of the command. If <0, it's a local message not meant to be returned somewhere
        returns: True if the command is consumed, otherwise false
        """
        # blank message is consumed here, no further processing takes place
        if len(msg)==0:
            return True

        tok     = msg.split(":")
        if len(tok)==0:
            self._logger.warning("OOPS: message from port {} yielded no tokens - '{}'".format(ingressPort, msg))
            return False

        cmd     = tok[0]
        try:
            if cmd=="$SEI":
                epicInpId       = int(tok[1])
                enable          = int(tok[2])
                lineIn          = self._getEpicLineIn(epicInpId)
                lineIn.enabled  = (enable!=0)
                uciSet(Uci(),"ae_ms","mixer","enable{}".format(epicInpId),"true" if enable else "false")
                with self._lock:
                    self._pRefreshInputs()
                    self._sendMsg("{}\r\n".format(msg), ingressPort)
                    v=-79
                    if(lineIn.enabled):
                        v=lineIn.volume
                    self._bcastMsgExcl("$CEI:{}:{}\r\n".format(epicInpId, enable))
                    self._bcastMsgExcl("$CVI:{}:{}\r\n".format(epicInpId, v))
                    return True
            elif cmd=="$QEI":
                epicInpId       = int(tok[1])
                lineIn          = self._getEpicLineIn(epicInpId)
                self._sendMsg("$QEI:{}:{}\r\n".format(epicInpId,1 if lineIn.enabled else 0), ingressPort)
                return True
            elif cmd=="$SXI":
                if tok[1].isdecimal():
                    epicInpId       = int(tok[1])
                    if epicInpId==0 or epicInpId in self.LINEIN_IDS():
                        self._inpExcReq = epicInpId
                        with self._lock:
                            self._logger.info("refreshing inputs AE CMD: {}".format(msg))
                            self._pRefreshInputs()
                            self._sendMsg("{}\r\n".format(msg), ingressPort)
                            self._bcastMsgExcl("$CXI:{}\r\n".format(lineId))
                return True
            elif cmd=="$SEO":
                lineId  = int(tok[1])
                enabled = int(tok[2])
                lineOut         = self._getEpicLineOut(lineId)
                lineOut.enabled = (enabled!=0)
                uciSet(Uci(),"ae_ms","volume","enable{}".format(lineId),"true" if enabled else "false")
                with self._lock:
                    self._pRefreshLineOut(lineId)
                    self._sendMsg("{}\r\n".format(msg), ingressPort)
                    v=-79
                    if(lineOut.enabled):
                        v=lineOut.volume
                    self._bcastMsgExcl("$CEO:{}:{}\r\n".format(lineId, enabled))
                    self._bcastMsgExcl("$CVO:{}:{}\r\n".format(lineId, v))
                    return True
            elif cmd=="$QEO":
                lineId  = int(tok[1])
                with self._lock:
                    lineOut = self._getEpicLineOut(lineId)
                    if lineOut is not None:
                        self._sendMsg("$QEO:{}:{}\r\n".format(lineId, 1 if lineOut.enabled else 0), ingressPort)
                        return True
                    else:
                        self._logger.warning("EPIC command is invalid (line out is NULL): '{}'".format(msg))
                        return False
            elif cmd=="$SVI":
                lineId  = int(tok[1])
                with self._lock:
                    lineIn          = self._getEpicLineIn(lineId)
                    lineIn.volume   = int(tok[2])
                    self._logger.info("refreshing inputs AE CMD: {}".format(msg))
                    self._pRefreshInputs()
                    uciSet(Uci(),"ae_ms","mixer","ch{}".format(lineId),tok[2])
                    self._sendMsg(msg, ingressPort)
                    self._bcastMsgExcl("$CVI:{}:{}\r\n".format(lineId, lineIn.volume))
                    return True
            elif cmd=="$QVI":
                lineId  = int(tok[1])
                with self._lock:
                    lineIn          = self._getEpicLineIn(lineId)
                    self._sendMsg("$QVI:{}:{}\r\n".format(lineId, lineIn.volume), ingressPort)
                    return True
            elif cmd=="$SVO":
                lineId  = int(tok[1])
                with self._lock:
                    lineOut         = self._getEpicLineOut(lineId)
                    lineOut.volume  = int(tok[2])
                    self._pRefreshLineOut(lineId)
                    self._sendMsg("{}\r\n".format(msg), ingressPort)
                    self._bcastMsgExcl("$CVO:{}:{}\r\n".format(lineId, lineOut.volume))
                uciSet(Uci(),"ae_ms","volume","out{}".format(lineId),str(lineOut.volume))
                return True
            elif cmd == "$IVO":
                lineId = int(tok[1])
                with self._lock:
                    lineOut         = self._getEpicLineOut(lineId)
                    if lineOut.volume<16:
                        lineOut.volume += 5
                        if lineOut.volume>16:
                            lineOut.volume  = 16
                        self._pRefreshLineOut(lineId)
                        self._sendMsg("{}\r\n".format(msg), ingressPort)
                        self._bcastMsgExcl("$CVO:{}:{}\r\n".format(lineId,lineOut.volume))
                        uciSet(Uci(), "ae_ms", "volume", "out{}".format(lineId), str(lineOut.volume))
                return True
            elif cmd == "$DVO":
                lineId = int(tok[1])
                with self._lock:
                    lineOut = self._getEpicLineOut(lineId)
                    if lineOut.volume>-79:
                        lineOut.volume -= 5
                        if lineOut.volume<-79:
                            lineOut.volume  = -79
                        self._pRefreshLineOut(lineId)
                        self._sendMsg("{}\r\n".format(msg), ingressPort)
                        self._bcastMsgExcl("$CVO:{}:{}\r\n".format(lineId, lineOut.volume))
                        uciSet(Uci(), "ae_ms", "volume", "out{}".format(lineId), str(lineOut.volume))
                return True
            elif cmd == "$IVI":
                lineId = int(tok[1])
                with self._lock:
                    if lineId==9:
                        self._ivi9(msg,ingressPort)
                    else:
                        lineIn      = self._getEpicLineIn(lineId)
                        if lineIn.volume<16:
                            lineIn.volume += 5
                            if lineIn.volume>16:
                                lineIn.volume  = 16
                            self._pRefreshInputs()
                            self._sendMsg("{}\r\n".format(msg), ingressPort)
                            self._bcastMsgExcl("$CVI:{}:{}\r\n".format(lineId,lineIn.volume))
                            uciSet(Uci(), "ae_ms", "mixer", "ch{}".format(lineId), str(lineIn.volume))
                return True
            elif cmd == "$DVI":
                lineId = int(tok[1])
                with self._lock:
                    if lineId==9:
                        self._dvi9(msg,ingressPort)
                    else:
                        lineIn = self._getEpicLineIn(lineId)
                        if lineIn.volume > -79:
                            lineIn.volume -= 5
                            if lineIn.volume < -79:
                                lineIn.volume = -79
                            self._pRefreshInputs()
                            self._sendMsg("{}\r\n".format(msg), ingressPort)
                            self._bcastMsgExcl("$CVI:{}:{}\r\n".format(lineId, lineIn.volume))
                            uciSet(Uci(), "ae_ms", "mixer", "ch{}".format(lineId), str(lineIn.volume))
                return True
            elif cmd=="$QVO":
                lineId  = int(tok[1])
                with self._lock:
                    lineOut = self._getEpicLineOut(lineId)
                    self._sendMsg("$QVO:{}:{}\r\n".format(lineId, lineOut.volume), ingressPort)
                    return True
            elif cmd=="$SPF":
                portId      = int(tok[1])
                if ingressPort<1 or ingressPort>6:
                    self._logger.error("invalid ingress port for $SPF: {}".format(portId))
                    return True

                egressList= list()
                # special case is forward to NULL, which simply clears the list.
                if tok[2]!="0":
                    portList = tok[2].split(",")
                    for p in portList:
                        if not p.isdigit():
                            self._logger.error("invalid egress port for $SPF:{} - {}".format(portId,p))
                            return True
                        v   = int(p)
                        if v==portId or v<1 or v>6:
                            self._logger.error("invalid egress port for $SPF:{} - {}".format(portId,p))
                            return True
                        egressList.append(v)
                with self._lock:
                    self._egressList[portId]    = egressList
                    self._logger.info("serial port {}: new egress list is {}".format(portId,egressList))
                    uciSet(Uci(),"ae_ms","serial","egress_list_{}".format(portId),tok[2])
                self._sendMsg("{}\r\n".format(msg), ingressPort)
                return True
            elif cmd=="$SSP":
                portList    = tok[1].split(",")
                payload     = bytes.fromhex(tok[2])
                for p in portList:
                    try:
                        self._sendBytes(payload, int(p))
                    except Exception as e:
                        self._logger.error("error sending {} to port {}".format(payload,p),exc_info=e)
                if ingressPort != 1:
                    self._sendMsg("{}\r\n".format(msg), ingressPort)
                return True
            elif cmd=="$CVU":
                if tok[1]=="3":
                    with self._lock:
                        self._ivi9(msg,ingressPort)
                return True
            elif cmd=="$CVD":
                if tok[1]=="3":
                    with self._lock:
                        self._dvi9(msg,ingressPort)
                return True
            elif cmd=="$SEQ":
                strBands    = tok[1:]
                if len(strBands)!=5:
                    self._logger.warning("invalid equalizer bands (need 5): {}".format(strBands))
                    return True
                bvalues     = list()
                for b in strBands:
                    if b.startswith("-"):
                        s   = -1
                        b2  = b[1:]
                    else:
                        s   = 1
                        b2  = b
                    if not b2.isdecimal():
                        self._logger.error("invalid band adjustment: {}".format(b))
                        return True
                    v       = s*int(b2)
                    if v<-15 or v>15:
                        self._logger.error("invalid band adjustment: {}".format(b))
                        return True
                    bvalues.append(v)
                self._eqBand[1] = bvalues[0]
                self._eqBand[2] = bvalues[1]
                self._eqBand[3] = bvalues[2]
                self._eqBand[4] = bvalues[3]
                self._eqBand[5] = bvalues[4]
                uci             = Uci()
                uciSet(uci, "ae_ms", "eq", "band1", strBands[0])
                uciSet(uci, "ae_ms", "eq", "band2", strBands[1])
                uciSet(uci, "ae_ms", "eq", "band3", strBands[2])
                uciSet(uci, "ae_ms", "eq", "band4", strBands[3])
                uciSet(uci, "ae_ms", "eq", "band5", strBands[4])
                self._updateEq()
                self._sendMsg(msg,ingressPort)
                return True
            elif cmd=="$QEQ":
                self._sendMsg("$QEQ:{}:{}:{}:{}:{}"
                              .format(self._eqBand[1],
                                      self._eqBand[2],
                                      self._eqBand[3],
                                      self._eqBand[4],
                                      self._eqBand[5]),ingressPort)
                return True
            elif cmd=="$SEE":
                valid  = True
                if tok[1]=="1":
                    self._eqEnable  = True
                elif tok[1]=="0":
                    self._eqEnable  = False
                else:
                    valid   = False

                if valid:
                    uciSet(Uci(),"ae_ms","eq","enable","true" if self._eqEnable else "false")
                    self._updateEq()
                    self._sendMsg(msg,ingressPort)
                return True
            elif cmd=="$QEE":
                self._sendMsg("$QEE:{}".format(1 if self._eqEnable else 0),ingressPort)
                return True
            elif cmd=="$SPB":
                if tok[1].isdecimal() and tok[2].isdecimal():
                    serialId    = int(tok[1])
                    if serialId<1 or serialId>5:
                        self._logger.error("invalid serial port id: {}".format(tok[1]))
                        return True
                    serialConfig= "{}:N:8:1".format(tok[2])
                    execCmd     = "#SSC:{}:{}".format(serialId,serialConfig)
                    uciSet(Uci(),"ae_ms","serial","config_{}".format(serialId),serialConfig)
                    self._tty.runMcuCmd(execCmd)
                    self._sendMsg("{}\r\n".format(msg), ingressPort)
                return True
            elif cmd == "$QPB":
                if tok[1].isdecimal():
                    serialId = int(tok[1])
                    if serialId < 1 or serialId > 5:
                        self._logger.error("invalid serial port id: {}".format(tok[1]))
                        return True
                    serialConfig= uciGet(Uci(), "ae_ms", "serial", "config_{}".format(serialId))
                    baud        = serialConfig.split(':')[0]
                    self._sendMsg("$QPB:{}:{}".format(serialId,baud), ingressPort)
                return True
            elif cmd=="$QPF":
                if tok[1].isdecimal():
                    portId      = int(tok[1])
                    if portId<1 or portId>6:
                        self._logger.error("invalid ingress port for $SPF: {}".format(portId))
                        return True

                    egressList  = self._egressList[portId]
                    if len(egressList)==0 or (len(egressList)==1 and egressList[0]==0):
                        self._sendMsg("$QPF:{}:0".format(portId), ingressPort)
                        return True
                    txtList = "{}".format(egressList[0])
                    for p in egressList[1:]:
                        txtList += ",{}".format(p)
                    self._sendMsg("$QPF:{}:{}".format(portId,txtList), ingressPort)
                return True
            elif cmd=="$QCC":
                uci         = Uci()
                reply       = ""
                reply      += "$SRS\r\n"
                for lineId, lineIn in self._lineIn.items():
                    reply      += "$SEI:{}:{}\r\n".format(lineId,1 if lineIn.enabled else 0)
                    reply      += "$SVI:{}:{}\r\n".format(lineId,lineIn.volume)
                    self._sendMsg(reply,ingressPort)
                    reply           = ""
                for lineId, lineOut in self._lineOut.items():
                    reply      += "$SEO:{}:{}\r\n".format(lineId,1 if lineOut.enabled else 0)
                    reply      += "$SVO:{}:{}\r\n".format(lineId,lineOut.volume)
                    self._sendMsg(reply,ingressPort)
                    reply           = ""
                reply      += "$SEE:{}\r\n".format(self._eqEnable)
                reply      += "$SEQ:{}:{}:{}:{}:{}\r\n".format(self._eqBand[1],
                                                               self._eqBand[2],
                                                               self._eqBand[3],
                                                               self._eqBand[4],
                                                               self._eqBand[5] )
                for serialId in range(1,7):
                    serialConfig    = uciGet(Uci(), "ae_ms", "serial", "config_{}".format(serialId))
                    baud            = serialConfig.split(':')[0]
                    reply          += "$SPB:{}:{}\r\n".format(serialId,baud)
                    egressList      = self._egressList[serialId]
                    if len(egressList) == 0 or (len(egressList) == 1 and egressList[0] == 0):
                        reply      += "$SPF:{}:0\r\n".format(serialId)
                    else:
                        txtList     = "{}".format(egressList[0])
                        for p in egressList[1:]:
                            txtList+= ",{}".format(p)
                        reply      += "$SPF:{}:{}\r\n".format(serialId, txtList)
                    self._sendMsg(reply,ingressPort)
                    reply           = ""
                reply      += "$SDI:4:MS700\r\n"
                reply      += "$SPT:2\r\n"
                reply      += "$SDF:3:14\r\n"
                self._sendMsg(reply,ingressPort)
                return True
            elif cmd.startswith('#'):
                # TODO: remove this branch for production
                # Direct access to the IO controller is for development only!!
                if ingressPort==6:
                    self._logger.info("NOTE: running command: {}".format(msg))
                    prefixes      = list()
                    if cmd=="#QIL":
                        prefixes.append("#QIL")
                    res , replies = self._tty.runMcuCmd(msg,prefixes)
                    reply         = ""
                    if len(replies)>0:
                        reply     = "{}\r\n".format(replies)
                    if res==0:
                        reply    += "OK\r\n"
                    else:
                        reply    += "ERROR\r\n"
                    self._sendMsg(reply, ingressPort)
                    return True
            else:
                return False
        except Exception as e:
            self._logger.error("OOPS: exception handling '{}' from port {}".format(msg, ingressPort), exc_info=e)
            # Consume the message because, for an exception to be thrown and
            # caught, the message command has to be understood, despite the
            # remaining payload containing some kind of error.
            return True



    def _acceptOneClient( self,skt:socket.socket ) -> socket.socket :
        rdlist, _, _ = select((skt,),(),(),0.001)
        if rdlist:
            try:
                (sktCon , clientAddr)  = skt.accept()
                sktCon.setblocking(False)
                sktCon.setsockopt(socket.SOL_SOCKET,socket.SO_KEEPALIVE,1)
                sktCon.setsockopt(socket.IPPROTO_TCP,socket.TCP_KEEPCNT,2)
                sktCon.setsockopt(socket.IPPROTO_TCP,socket.TCP_KEEPIDLE,5)
                sktCon.setsockopt(socket.IPPROTO_TCP,socket.TCP_KEEPINTVL,5)
                logMsg                 = "client connected from {}".format(clientAddr)
                self._logger.info(logMsg)
                syslog.syslog(logMsg)
                return sktCon
            except Exception as e:
                self._logger.error("Unexpected exception: {}".format(e))
                return None

    def qccResponse(self):
        self._logger.info("Re-sending audio settings")
        
        with self._lock:
            self._pRefreshInputs()
            self._pRefreshLineOut()

    def handleIoMcuMsg(self, payload:bytes) -> None:
        """
        Handle one serial message from the MCU
        msg - serial message, 
              already stripped of terminating characters
        returns: TBD
        """
        if len(payload)==0:
            return

        try:
            msg = payload.decode('ascii')
        except Exception as e:
            # if payloads fails conversion to string, don't do anything here.
            self._logger.error("unable to convert MCU msg to string: {}".format(payload),exc_info=e)
            return

        #self._logger.info("TTY msg: '{}'".format(msg))

        comps   = msg.split(":")
        if len(comps)==0:
            self._logger.warning("discarding phony message: '{}'".format(msg))
            return

        cmd     = comps[0]
        args    = list()
        if len(comps)>1:
            args= comps[1:]

        if cmd == "#CPL":
            comps = msg.split(":")
            pwrt = float(comps[4])
            pwavg = float(comps[3])
            pwmin = float(comps[2])
            pwmax = float(comps[1])
            self.pwRT = pwrt
            self.pwAcc1m = self.pwAcc1m + pwavg
            self.count1m = self.count1m + 1
            self.pwAcc1h = self.pwAcc1h + pwavg
            self.count1h = self.count1h + 1
            if self.pwMax1m < pwmax:
                self.pwMax1m = pwmax
            if self.pwMin1m > pwmin:
                self.pwMin1m = pwmin
            if self.pwMax1h < pwmax:
                self.pwMax1h = pwmax
            
            t = time.time()
            if t - self.lastTime1m >= 60:
                self.pwAvg1m = self.pwAcc1m / self.count1m
                self.pwAcc1m = pwrt
                self.count1m = 1
                self.pwMax1m = pwrt
                self.pwMin1m = pwrt
                self.lastTime1m = t
                
            if t - self.lastTime1h >= 3600:
                self.pwAvg1h = self.pwAcc1h / self.count1h
                self.pwAcc1h = pwrt
                self.count1h = 1
                self.pwMax1h = pwrt
                self.lastTime1h = t
                
        elif cmd=="#CIL":
            if len(args)<2:
                self._logger.warning("Invalid #CIS message. need 2 args, have {}".format(len(args)))
                return
            gdPin   = int(args[0])
            evt     = InOut.PRESSED
            if args[1]=="1":
                evt = InOut.RELEASED
            
            if gdPin==self._ipin1.gdPin:
                self._ipin1.events.append(evt)
            elif gdPin==self._ipin2.gdPin:
                self._ipin2.events.append(evt)
            elif gdPin==self._ipin3.gdPin:
                self._ipin3.events.append(evt)
            elif gdPin==self._ipin4.gdPin:
                self._ipin4.events.append(evt)
        elif cmd=="#QCC":
            self.qccResponse()
        elif cmd=="#QDI":   # This shouldn't be here. To be fixed with the next GD fw version. The GD should only return the fw version - not the #QDI command.
            self.deviceInfo['gdVer'] = comps[1]
            self._sendMsg("{}\r\n".format(msg), 6)
        else:
            self._logger.warning("Unknown GD message: '{}'".format(msg))
    
    def consumptionAvg1m(self) -> str:
        v = self.pwAvg1m
        if v == 0 and self.count1m > 0:
            v = self.pwAcc1m / self.count1m
        v = round(v * 24, 2)
        return str(v)
    
    def consumptionAvg1h(self) -> str:
        v = self.pwAvg1h
        if v == 0 and self.count1h > 0:
            v = self.pwAcc1h / self.count1h
        v = round(v * 24, 2)
        return str(v)
    
    def consumptionMax1m(self) -> str:
        return str(round(self.pwMax1m * 24, 2))
    
    def consumptionMax1h(self) -> str:
        return str(round(self.pwMax1h * 24, 2))
    
    def consumptionMin1m(self) -> str:
        return str(round(self.pwMin1m * 24, 2))
    
    def consumptionRT(self) -> str:
        return str(round(self.pwRT * 24, 2))
    
    def _pRequireLock(self) -> None:
        if not self._lock.locked():
            raise Exception("Lock needs to be aquired!")

    def _pRefreshInputs(self) -> None:
        self._pRequireLock()
        exclusive   = self._inpExcReq
        chRoom      = dict()
        chRoom["mute"]  = False
        chRoom["gain"]  = 0.0
        chStreams       = dict()

        for lineId, lineIn in self._lineIn.items():
            if exclusive!=0 and exclusive!=lineId:
                gain    = -79
                mute    = True
            else:
                mute    = not lineIn.enabled
                gain    = lineIn.volume

            if(not lineIn.enabled):
                gain=-79

            if lineId==self.LINEIN_STREAMS():
                chStreams["mute"]   = mute
                chStreams["gain"]   = gain
            else:
                self._tty.runMcuCmd("#SIV:{}:{}".format(lineId,gain))
        try:
            dspMixerReq             = dict()
            dspMixerReq["room"]     = chRoom
            dspMixerReq["streams"]  = chStreams
            r = HttpMethod.patch("http://127.0.0.1:18088/api/v1/media-control/mixer", dspMixerReq)
            self._logger.debug("result from PATCH mixer: {}".format(r))
        except Exception as e:
            self._logger.error("Exception setting mute on DSP:", exc_info=e)

    def _pConfigureMcuLineOut(self,lineId:int, gain:int, enable:bool):
        cmd = "#SLV:{}:{}".format(lineId, gain if enable else -79)
        self._tty.runMcuCmd(cmd)

    def _pRefreshLineOut(self, lineId:int = -1) -> None:
        """
        Refresh one or all LineOut objects
        id: 1 to 6 for specific output, or -1 for all (default)
        Requires a lock...
        """
        self._pRequireLock()
        if lineId != -1:
            lineOut = self._getEpicLineOut(lineId)
            if lineId==0:
                self.setMainVolume(lineOut.volume if lineOut.enabled else -79)
            else:
                self._pConfigureMcuLineOut(lineId,lineOut.volume,lineOut.enabled)
        else:
            for lineId, lineOut in self._lineOut.items():
                lineOut = self._getEpicLineOut(lineId)
                if lineId == 0:
                    self.setMainVolume(lineOut.volume if lineOut.enabled else -79 )
                else:
                    self._pConfigureMcuLineOut(lineId, lineOut.volume, lineOut.enabled)

    def _readFromTcp(self,skt:socket.socket) -> str:
        if len(self._tcpRxQueue)==0:
            rdlist, _, xList = select((skt,), (), (skt,), 0.05)
            if xList is not None and skt in xList:
                raise Exception("client connection in error state")

            if rdlist is None or skt not in rdlist:
                return ""

            b   = skt.recv(128)
            if len(b)==0:
                # empty read with ready socket = disconnected
                raise Exception("empty read while ready")

            self._tcpRxQueue.extend(b)
        while len(self._tcpRxQueue)>0:
            c   = self._tcpRxQueue.pop(0)
            if c>127:
                continue
                
            self._tcpBuf.append(c)
            if c==ord('\n'):
                line        = ""
                try:
                    # lines may have invalid characters. They'll be discarded in this case.
                    line    = self._tcpBuf.decode('ascii')
                except Exception as ex:
                    self._tcpBuf.clear()
                    self._logger.error("exception decoding input from socket: ",exc_info=ex)
                finally:
                    self._tcpBuf.clear()
                    return line.strip()
        return ""

    def _sendBytes(self, msg:bytes, port:int) -> None:
        """
        Sends one raw message (bytes) from the specified port
        msg: payload to send
        port: if the port
        """
        if len(msg)==0:
            return
        try:
            if port==6 and self._sktCon is not None:
                self._sktCon.send(msg)
            elif port>0 and port<6:
                self._tty.sendBytes(msg,port)
        except Exception as e:
            self._logger.error("failed to send/fwd '{}' from port {}".format(msg,port),exc_info=e)

    def _sendMsg(self, msg:str, port:int ) -> None:
        """
        Sends one text message through the specified port
        msg: the string to send. If msg does not end with neither CR nor LF,
             a terminating CRLF is appended automatically
        port: the egress port (1 to 6)
        """
        if len(msg)==0:
            return

        if not msg.endswith("\n") and not msg.endswith("\r"):
            msg += "\r\n"

        try:
            if port==6 and self._sktCon is not None:
                self._sktCon.send(msg.encode("ascii"))
            elif port>0 and port<6:
                self._tty.sendMessage(msg,port)
        except Exception as e:
            self._logger.error("failed to send/fwd '{}' from port {}".format(msg,port),exc_info=e)


    def _bcastMsgIncl(self, msg:str, ports: List[int]) -> None:
        if len(msg)==0:
            return
        msg += "\r\n"
        for port in ports:
            self._sendMsg(msg,port)

    def _bcastMsgExcl(self, msg:str, ports: List[int]=[]) -> None:
        if len(msg)==0:
            return
        msg += "\r\n"
        for port in range(1,7):
            if port in ports:
                continue
            self._sendMsg(msg,port)

    def _eventLoop(self):
        res , replies = self._tty.runMcuCmd("#QDI:0", ["#QDI"])
        if res==0:
            gdV = "".join(replies)[7:]
            self.deviceInfo['gdVer'] = gdV
            self._logger.info("GD firmware version: {}".format(gdV))
            
        res , replies = self._tty.runMcuCmd("#QDI:1", ["#QDI"])
        if res==0:
            self._logger.info("BIC version: {}".format("".join(replies)[7:]))
        
        skt     = None
        if self._gwMode==SerialGw.SERVER:
            logMsg  = "Serial GW SERVER: listening on port {}".format(self._gwPort)
            self._logger.info(logMsg)
            syslog.syslog(logMsg)
            skt = socket.socket(socket.AF_INET,socket.SOCK_STREAM)
            skt.setsockopt(socket.SOL_SOCKET,socket.SO_REUSEADDR,1)
            skt.bind(("0.0.0.0",self._gwPort))
            skt.listen(1)

        with self._lock:
            self._tty.runMcuCmd("#SPC:100:0:0:0")
            # Amplifier is unmuted by the DSP when there is no silence
            self._tty.runMcuCmd("#SAM:1")
            self._tty.runMcuCmd("#STC:0")
            self._tty.runMcuCmd("#SOL:16:1")    # activate power supply to Peripherals (WPA...)
            self._tty.runMcuCmd("#SOL:5:1")     # activate power supply to XD receiver

        # do not restart the DSP on startup
        self.reloadSettings()

        self._sktCon      = None
        while not self._stopGw:
            # actively refuse concurrent connections while the current
            # client is connected.
            if skt is not None:
                s   = self._acceptOneClient(skt)
                if s is not None:
                    if self._sktCon is None:
                        self._sktCon    = s
                    else:
                        print("Serial GW refused connection")
                        self._logger.info("Serial GW refused connection")
                        s.close()

            rxMessages          = list()
            if self._sktCon is not None:
                try:
                    line        = self._readFromTcp(self._sktCon)
                    if len(line)>0:
                        rxMessages.append( (6,bytearray(line.strip(),'ascii')) )
                except Exception as ex:
                    self._logger.info("client disconnected: {}".format(str(ex)))
                    self._sktCon.close()
                    self._tcpBuf.clear()
                    self._sktCon    = None

            serialMsgs  = self._tty.readMessages()
            if len(serialMsgs)>0:
                rxMessages.extend(serialMsgs)

            for port, payload in rxMessages:
                try:
                    if port==0:
                        self.handleIoMcuMsg(payload)
                    elif port>0 and port<7:
                        consumed    = False
                        try:
                            msg     = payload.decode('ascii').strip()
                            self._logger.warning("handling '{}' from port {}...".format(msg,port))
                            consumed= self.handleAeCommand(msg, port)
                            self._logger.warning("'{}' from port {} was {}consumed...".format(msg, port,"" if consumed else "NOT "))
                            if not consumed:
                                # message not consumed ==> basic routing
                                self._logger.warning("routing '{}': {} --> {}".format(msg,port,self._egressList[port]))
                                self._bcastMsgIncl(msg, self._egressList[port])
                        except:
                            pass
                        
                        if port!=6 and not consumed and self._sktCon is not None:
                            CSP_enabled = uciGet(Uci(),"ae_ms","serial","csp_{}".format(port))
                            if CSP_enabled=="true":
                                #generate a $CSP with the payload
                                notification = "$CSP:{}:{}\r\n".format(port,payload.hex().upper())
                                self._sktCon.send(notification.encode('ascii'))
                    else:
                        raise IndexError("message '{}' from unexpected port: {}".format(payload, port))

                except Exception as ex:
                    self._logger.error("Exception handling tty line: '{}'".format(payload), exc_info=ex)

            with self._lock:
                if self._displayTimeout!=0 and time.monotonic()>self._displayTimeout:
                    self._logger.warning(">>>>> IMPLEMENT DEACTIVATE DISPLAY")
                    self._displayTimeout    = 0
            
            time.sleep(0.005)

        self._logger.info("gateway worker exiting.")


    def shutdown(self):
        """
        Shutdown operation of the serial GW and disconnect
        from the serial port
        """
        self._tty.shutdown()
        self._stopGw = True
        self._logger.info("waiting for worker thread to stop")
        self._workerThread.join()

    def setExclusiveInput(self, inpId:int) -> None:
        """
        Set an exclusive input (1..6) or disable it with 0
        """
        if inpId!=0 and inpId not in self.LINEIN_IDS():
            raise Exception("Invalid input specifier: {}".format(inpId))

        self._logger.info("EXCLUSIVE INPUT: {} --> {}".format(self._inpExcReq,inpId))

        with self._lock:
            self._inpExcReq = inpId
            self._logger.debug("refreshing inputs @ SXI:{}".format(inpId))
            self._pRefreshInputs()


    def writeSideChannel(self,data:str) -> None :
        """
        Write data to the sidechannel of the serial port
        data - string to write into the serial port, regardless of
               the operation mode.
        """
        self._tty.writeSideChannel(data)

    def setMainVolume(self, volume: int) -> None:
        """
        Sets the main volume output
        volume - the volume in dB, from -79 to +16
        """
        logger      = logging.getLogger("device")

        logger.info(f'Configured main volume: {volume}dB')

        uci = Uci()
        imp = uciGet(uci, "ae_ms", "speaker", "impedance")
        clippValue = self._getLineOutAttenuation(imp)

        body = { 'value_dB' : volume, 'clipLevel_dB': clippValue}
        logger.info("setting main volume: {}".format(body))
        try:
            r = HttpMethod.put("http://127.0.0.1:18088/api/v1/media-control/main-volume",body)
            logger.debug("result from PUT: {}".format(r))
        except Exception as e:
            logger.error("exception setting main volumere: ",exc_info=e)


    def _setMixerParam(self, param, value):
        p = subprocess.Popen("amixer -q -c 2 -- sset " + "\'" + param + "\'" + " \'" + value + "\'", shell=True)
        p.wait()

    def _updateEq(self) -> None:
        logger      = logging.getLogger("device")
        body = dict()
        body["enable"]  = self._eqEnable
        body["bands"]   = [
            self._eqBand[1] ,
            self._eqBand[2] ,
            self._eqBand[3] ,
            self._eqBand[4] ,
            self._eqBand[5]
            ]
        logger.info("setting EQ bands on DSP: {}".format(self._eqBand))
        try:
            r = HttpMethod.put("http://127.0.0.1:18088/api/v1/media-control/amp-eq",body)
            logger.debug("result from PUT: {}".format(r))
        except Exception as e:
            logger.error("exception reaching API: ",exc_info=e)


    def _clearIdPattern(self):
        with self._lock:
            self._idModeTmr = None
            self._tty.sendMessage("$RLD\r\n",5)

    def identifyXd(self) -> None:
        """
        Puts the device into identification pattern: the XD LEDs are put into
        a predefined mode.
        timeout - time in seconds, to restore normal operation
        """
        if self._idModeTmr is None:
            self._tty.sendMessage("$SLD:0:2\r\n$SLD:1:1\r\n$SLD:2:0\r\n",5)
        else:
            self._idModeTmr.cancel()

        self._idModeTmr = Timer(30,self._clearIdPattern)
        self._idModeTmr.start()

    def enableSSH(self, status: bool) -> None:
        if status:
            subprocess.Popen(["/etc/init.d/dropbear", "start"])
        else:
            subprocess.Popen(["/etc/init.d/dropbear", "stop"])
            subprocess.Popen(["killall", "dropbear"])
    
    def activateDisplay(self) -> None:
        with self._lock:
            if self._displayTimeout==0:
                self._logger.warning(">>>> IMPLEMENT ACTIVATE DISPLAY!!!")
            self._displayTimeout    = time.monotonic() + 60

    def muteAudioOutputs(self) -> None:
        if self._muteStatus is not None and self._muteStatus:
            return
        self._logger.info("Mutting amplifier")
        self._tty.runMcuCmd("#SBL:1:-79:-79:-79:-79")
        self._muteStatus = True

    def unmuteAudioOutputs(self) -> None:
        if self._muteStatus is not None and not self._muteStatus:
            return
        self._logger.info("Unmutting amplifier")
        msg="#SBL:0:"
        msg+=str(self._getEpicLineOut(1).volume)+":"
        msg+=str(self._getEpicLineOut(2).volume)+":"
        msg+=str(self._getEpicLineOut(3).volume)+":"
        msg+=str(self._getEpicLineOut(4).volume)
        self._tty.runMcuCmd(msg)
        self._muteStatus = False

