#!/usr/bin/env python3

import sys
import logging
import threading
import time
import serial
import struct
import json
import subprocess
import urllib.request
from barix.system import get_eth_mac
from barix.io.leds import IpamLed

logging.basicConfig(stream=sys.stdout)

class IoController:
	DISPLAY_COLOR_RED = 'R'
	DISPLAY_COLOR_GREEN = 'G'
	DISPLAY_COLOR_BLUE = 'B'
	DISPLAY_COLOR_YELLOW = 'Y'
	DISPLAY_COLOR_ORANGE = 'O'
	DISPLAY_COLOR_WHITE = 'W'
	DISPLAY_COLOR_BLACK = 'K'

	def __init__(self, port: str):
		self._logger = logging.getLogger("gd")
		self._logger.setLevel(logging.ERROR)
		self._lock = threading.Lock()

		self._stop = False
		self._tty = serial.Serial()
		self._tty.port = port
		self._tty.baudrate = 115200
		self._tty.bytesize = serial.EIGHTBITS
		self._tty.parity = serial.PARITY_NONE
		self._tty.stopbits = serial.STOPBITS_ONE
		self._tty.xonxoff = False
		self._tty.rtscts = False
		self._tty.dsrdtr = False
		self._tty.rs485_mode = None
		self._tty.timeout = 0.025
		self._tty.open()

		# TX queue, one per port...
		#        self._txQueue       = dict()
		#        for i in range(1,6):
		#            self._txQueue[i]= list()

		self._rdBuf = ""
		self._ttyRxMsgQueue = list()
		self._ttyRxThread = threading.Thread(target=self._ttyEventLoop)
		self._ttyRxThread.start()

	def _pRequireLock(self) -> None:
		if not self._lock.locked():
			raise Exception("Lock needs to be aquired!")

	def readMessages(self):
		"""
		Read all queued incoming messages
		Returns a list where each entry is a tuple with first element the
		port number of the message and the second element the actual message,
		as a byte array, for support of non-ascii payloads
		"""
		with self._lock:
			if len(self._ttyRxMsgQueue) > 0:
				msgs = list()
				rawMsgs = self._ttyRxMsgQueue.copy()
				self._ttyRxMsgQueue.clear()
				for entry in rawMsgs:
					try:
						if entry.startswith("#CSP:"):
							tokens = entry.split(":")
							msgs.append((int(tokens[1]), bytearray.fromhex(tokens[2])))
						else:
							msgs.append((0, bytearray(entry, 'ascii')))
					except Exception as e:
						self._logger.error("Error handling raw message: '{}'".format(entry), exc_info=e)
				return msgs
			return list()

	def sendBytes(self, msg: bytes, port: int):
		cmd = "#SSP:{}:{}".format(port, msg.hex().upper())
		self._logger.info("TX -> {}: '{}'".format(port, msg))
		self.runMcuCmd(cmd)

	def sendMessage(self, msg: str, port: int):
		cmd = "#SSP:{}:{}".format(port, msg.encode("ascii").hex().upper())
		self._logger.info("TX -> {}: '{}'".format(port, msg))
		self.runMcuCmd(cmd)

	def runMcuCmd(self, cmd: str, prefixes=[], timeout: int = 200):
		"""
		Runs an MCU command
		cmd - string with the command (\\r\\n is added automatically)
		timeout - timeout, in milliseconds (default: 200)
		prefixes - list of reply prefixes that will be used to detect
					non-default replies
		returns a tuple with:
			first element (int): 0 on success, failure codes TBD
			second element (str): the response to the command
		Must be run in a critical region
		"""
		with self._lock:
			self._logger.info("GD COMMAND TX: {}".format(cmd))
			expires = time.monotonic_ns() + timeout * 1000000

			replies = list()
			try:
				txdata = (cmd + "\r\n").encode("utf-8")
				self._tty.write(txdata)
				while time.monotonic_ns() < expires:
					self._logger.debug("reading line...")
					ttyLines = self._pReadLinesFromTty()
					while len(ttyLines) > 0:
						line = ttyLines.pop(0)
						if line == "OK":
							self._logger.debug("GD COMMAND '{}' --> OK, {} replies".format(cmd, len(replies)))
							if len(ttyLines) > 0:
								self._logger.debug("saving for later: " + ' '.join(ttyLines))
								self._ttyRxMsgQueue.extend(ttyLines)
							return (0, replies)

						if line.startswith("ERR"):
							self._logger.info("GD COMMAND '{}' --> ERROR".format(cmd))
							if len(ttyLines) > 0:
								self._logger.info("saving for later: " + ' '.join(ttyLines))
								self._ttyRxMsgQueue.extend(ttyLines)
							self._ttyRxMsgQueue.extend(ttyLines)
							return (1, replies)

						saveLine = True
						for prefix in prefixes:
							if line.startswith(prefix):
								self._logger.debug("GD COMMAND '{}' --> add reply '{}'".format(cmd, line))
								replies.append(line)
								saveLine = False
								break
						if saveLine:
							self._logger.debug("saving for later: '{}'".format(line))
							self._ttyRxMsgQueue.append(line)
				self._logger.info("GD COMMAND '{}' --> TIMEOUT".format(cmd))
				return (2, "")
			except Exception as e:
				self._logger.error("Exception running GD command: ", exc_info=e)
				return (-1, None)

	def _pReadLinesFromTty(self):
		"""
		Read available lines from the tty port and returns a list of whole lines.
		Any incomplete lines are kept in buffer for the next call
		"""
		self._pRequireLock()
		lines = list()
		while not self._stop:
			try:
				rxBytes = list(self._tty.read(64))
			except Exception as ex:
				if not self._stop:
					self._logger.error("Exception reading tty:", exc_info=ex)
				return lines

			if len(rxBytes) == 0:
				# print(".", end='', flush=True)
				return lines

			for b in rxBytes:
				c = chr(b)
				self._rdBuf += c
				if c == '\n':
					line = self._rdBuf.strip()
					# self._logger.debug("TTY RX: '{}'".format(line))
					lines.append(line)
					self._rdBuf = ""
		return []

	def _ttyEventLoop(self):
		while not self._stop:
			with self._lock:
				self._ttyRxMsgQueue.extend(self._pReadLinesFromTty())

	def shutdown(self):
		"""
		Shutdown operation of the serial GW and disconnect
		from the serial port
		"""
		self._stop = True
		self._tty.close()
		self._tty.cancel_read()
		self._logger.info("waiting for worker thread to stop")
		self._ttyRxThread.join()

	def writeSideChannel(self, data: str) -> int:
		"""
		Write data to the sidechannel of the serial port
		data - string to write into the serial port, regardless of
			   the operation mode.
		"""
		with self._lock:
			return self._tty.write(data.encode("utf-8"))

	def turnOnDisplay(self):
		self.runMcuCmd('#SDS:1')

	def setDisplayColor(self, font:bool, color:str):
		font=int(font)
		self.runMcuCmd(f'#SDC:{font}:{color}')
	
	def setDisplayText(self, line:int, text:str):
		text="".join("{:02x}".format(ord(c)) for c in text).upper()
		self.runMcuCmd(f'#SDT:{line}:{text}')


if __name__=="__main__":
	if len(sys.argv) != 2:
		print('Invalid arguments!')
		sys.exit(1)

	gdCtl = IoController("/dev/ttyACM0")
	alive, _ = gdCtl.runMcuCmd('#IIA')
	while alive != 0:
		time.sleep(1.0)
		alive, _ = gdCtl.runMcuCmd('#IIA')
	gdCtl.turnOnDisplay()
	gdCtl.setDisplayColor(False, IoController.DISPLAY_COLOR_GREEN)
	gdCtl.setDisplayColor(True, IoController.DISPLAY_COLOR_BLACK)
	gdCtl.setDisplayText(1, sys.argv[1])
	mac = get_eth_mac().upper()
	gdCtl.setDisplayText(2, mac[:9])
	gdCtl.setDisplayText(3, mac[9:])

	# blink both LEDs
	led1 = IpamLed("led1")
	led1.set("green", True, (100, 100))

	output = subprocess.check_output("/usr/bin/qiba-spi-get-info").decode()
	factory_info = json.loads(output)

	with open("/mnt/shadow/config.json", "r") as f:
		config = json.load(f)

	sticker_info = {
			"device": "MS700",
			"mac": get_eth_mac(),
			"prod_date": factory_info["HW_DEVICE"]["Production_Date"],
			"regid": config["registrationId"],
			"webui_user": "admin",
			"webui_pwd": "admin"
	}

	# wait for reset button to print sticker
	in_file = open("/dev/input/event0", "rb")
	FORMAT = 'llHHI'
	EVENT_SIZE = struct.calcsize(FORMAT)
	while True:
		event = in_file.read(EVENT_SIZE)
		(tv_sec, tv_usec, type, code, value) = struct.unpack(FORMAT, event)
		if code == 256:
			if value == 1:
				# send message to print sticker
				try:
					req = urllib.request.Request(url="http://192.168.2.221:8899/print",
						data=json.dumps(sticker_info).encode(),
						headers={"Content-Type": "application/json"},
						method="PUT")
					urllib.request.urlopen(req, timeout=5)
				except Exception as exc:
					print(f"{exc}")
			else:
				time.sleep(1.0)

	gdCtl.shutdown()
